import os
import glob
from pele_platform.Allosteric.cluster import cluster_best_structures
from pele_platform.PPI.simulation_launcher import launch_simulation
from pele_platform.PPI.preparation import prepare_structure
from pele_platform.Utilities.Helpers.helpers import cd, is_repited, is_last
import pele_platform.Utilities.Parameters.pele_env as pv
import pele_platform.Adaptive.simulation as si


def run_ppi(parsed_yaml: dict) -> (pv.EnviroBuilder, pv.EnviroBuilder):


    # Let user choose working folder
    original_dir = os.path.abspath(os.getcwd())
    working_folder = os.path.abspath("{}_Pele".format(parsed_yaml.residue))
    if not parsed_yaml.folder:
        working_folder = is_repited(working_folder) if not parsed_yaml.adaptive_restart else is_last(working_folder)
    else:
        working_folder = os.path.abspath(parsed_yaml.folder)

    # Set main folder
    parsed_yaml.folder = os.path.join(working_folder, "1_interface_exploration")

    # Check n_waters before launching the simulation
    #water_checker(parsed_yaml)

    # get arguments from input.yaml
    n_waters = parsed_yaml.n_waters
    parsed_yaml.n_waters = None
    protein_file = parsed_yaml.system
    chain = parsed_yaml.protein
    ligand_pdb = parsed_yaml.ligand_pdb

    # no waters in the first simulation
    parsed_yaml.water_arg = None 
   
    # remove chains except for "protein" flag
    protein_file = prepare_structure(protein_file, ligand_pdb, chain, True)
    parsed_yaml.system = protein_file

    # start simulation 1 - induced fit
    parsed_yaml.induced_fit_exhaustive = True
    simulation1 = si.run_adaptive(parsed_yaml)
    simulation1_path = os.path.join(simulation1.pele_dir, simulation1.output)

    # cluster best structures
    if not parsed_yaml.debug:
        with cd(simulation1_path):
            cluster_best_structures("5", n_components=simulation1.n_components,
                residue=simulation1.residue, topology=simulation1.topology,
                directory=working_folder, logger=simulation1.logger)
    
    # adjust original input.yaml
    if not parsed_yaml.skip_refinement:
        parsed_yaml.system = os.path.join(working_folder, "refinement_input/*.pdb")
        parsed_yaml.folder = os.path.join(working_folder, "2_refinement_simulation")
        parsed_yaml.induced_fit_exhaustive = None
        parsed_yaml.ppi = None
        parsed_yaml.poses = None
        parsed_yaml.rescoring = True
        del parsed_yaml.water_arg
        # Set waters ony if specified by user
        if n_waters != 0:
            parsed_yaml.waters = "all_waters"
            parsed_yaml.n_waters = n_waters
        else:
            parsed_yaml.waters = None
            parsed_yaml.n_waters = n_waters
        parsed_yaml.adaptive_restart = False
        if not parsed_yaml.test:
            parsed_yaml.iterations = 1
            parsed_yaml.steps = 100
        parsed_yaml.box_center = simulation1.box_center
        parsed_yaml.box_radius = 100  # We should have a look at how to set no box but at the moment super big

        # start simulation 2 - minimisation
        with cd(original_dir):
            if not parsed_yaml.debug:
                simulation2 = launch_simulation(parsed_yaml)
            else:
                simulation2 = None
    else:
        simulation2 = None
    return simulation1, simulation2



import os
import shutil
import logging
import pele_platform.constants.constants as cs
import pele_platform.features.adaptive as adfs
import pele_platform.features.frag as frfs
import pele_platform.Utilities.Helpers.helpers as hp
from pele_platform.Utilities.Parameters.SimulationParams import simulation_params
from pele_platform.Utilities.Parameters.SimulationFolders import simulation_folders


class EnviroBuilder(simulation_params.SimulationParams, simulation_folders.SimulationPaths):
    """
        Base class where the needed pele environment
        is build by creating folders and files
    """
    def __init__(self):
        pass

    def build_adaptive_variables(self, args):
        #DEFINE MAIN PATH
        pele_dir = os.path.abspath("{}_Pele".format(args.residue))
        if not args.folder:
            self.pele_dir = hp.is_repited(pele_dir) if args.restart in cs.FIRST_RESTART else hp.is_last(pele_dir)
            self.pele_dir = hp.is_repited(pele_dir) if not args.adaptive_restart else hp.is_last(pele_dir)
        else:
            self.pele_dir = os.path.abspath(args.folder)
        #####Define default variables, files and folder "HIDING VARIABLES " --> CHANGE#########
        for key, value in adfs.retrieve_software_settings(args, self.pele_dir).items():
            setattr(self, key, value)
        #####Initialize all variables by combining default and user input######
        simulation_params.SimulationParams.__init__(self, args)
        simulation_folders.SimulationPaths.__init__(self, args)
        for key, value in adfs.retrieve_software_settings(args, self.pele_dir).items():
            setattr(self, key, value)

    def build_frag_variables(self, args):
        # Define default variables, files and folder "HIDING VARIABLES " --> CHANGE
        for key, value in frfs.retrieve_software_settings(args).items():
            setattr(self, key, value)
        # Initialize all variables by combining default and user input
        simulation_params.SimulationParams.__init__(self, args)
        for key, value in frfs.retrieve_software_settings(args).items():
            setattr(self, key, value)
        self.create_logger(".")

    def create_files_and_folders(self):
        if not self.adaptive_restart:
            self.create_folders()
            self.create_files()
            self.create_logger()
        else:
            self.create_logger()

    def create_folders(self):
        """
            Create pele folders
        """
        for folder in self.folders:
            hp.create_dir(self.pele_dir, folder)

    def create_files(self):
        """
            Copy templates
        """

        # Actions
        for file, destination_name in zip(self.files, self.file_names):
            shutil.copy(file, os.path.join(self.pele_dir, destination_name))

    def create_logger(self, directory=None):
        directory = directory if directory else self.pele_dir
        log_name = os.path.join(directory, "{}.log".format(self.residue))
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        self.logger.propagate = False
        formatter = logging.Formatter("%(asctime)s:%(levelname)s:%(message)s")
        if not self.adaptive_restart:
            file_handler = logging.FileHandler(log_name, mode='w')
        else:
            file_handler = logging.FileHandler(log_name, mode='a')
        file_handler.setFormatter(formatter)
        self.logger.addHandler(file_handler)



from dataclasses import dataclass
import os
import pele_platform.Errors.custom_errors as ce

@dataclass
class EnvVariable():

    name: str
    variable: str
    default: str
    flag: str
    env_var: str

    def is_valid(self) -> None:
        #Check if environment variable is correctly set
        if self.variable:
            if os.path.exists(self.variable):
                return True
        elif self.default:
           if os.path.exists(self.default):
               return True
        raise ce.EnvVariableNotFound("{} not found. \n\
1) If you have the standard installation export the environment variable by doing:\n\
\t{}.\n\
2) If the previous does not work define the location of {} with the next flag in the input.yaml:\n\
\t{}".format(self.name, self.env_var, self.name, self.flag))




import os
from dataclasses import dataclass
import subprocess
import pele_platform.Utilities.Helpers.helpers as hp
import pele_platform.Utilities.Parameters.pele_env as pv
import pele_platform.Utilities.Helpers.plop_launcher as plop
import pele_platform.Errors.custom_errors as ce

@dataclass
class LigandParametrization:

    '''
    Base class to generate ligand parameters
    1) Generate forcefield and rotamers from ligand
    2) Copy user's external files
    '''

    env: pv.EnviroBuilder

    def generate(self) -> None:
        #Generate ligand forcefield parameters
        self.env.logger.info("Creating template for residue {}".format(self.env.residue))
        self.create_ligand_parameters()
        self.copy_external_parameters()
        self.env.logger.info("Template {}z created\n\n".format(self.env.residue.lower()))

    def create_ligand_parameters(self) -> None:
        #create ligz template files
        with hp.cd(self.env.pele_dir):
            plop.parametrize_miss_residues(self.env)

    def copy_external_parameters(self) -> None:
        #copy user's files
        self.copy_ligand_forcefield_file()
        self.copy_ligand_rotamer_file()

    def copy_ligand_forcefield_file(self) -> None:
        #copy user's ligz template files
        for template_file in self.env.external_template:
            if not os.path.exists(template_file):
                raise ce.TemplateFileNotFound(f"File {template_file} not found")
            cmd_to_move_template = "cp {} {}".format(template_file,  self.env.template_folder)
            subprocess.call(cmd_to_move_template.split())

    def copy_ligand_rotamer_file(self):
        #copy user's .rot.assign files
        for rotamer_file in self.env.external_rotamers:
            if not os.path.exists(rotamer_file):
                raise ce.RotamersFileNotFound(f"File {rotamer_file} not found")
            cmd_to_move_rotamer_file = "cp {} {}".format(rotamer_file, self.env.rotamers_folder)
            subprocess.call(cmd_to_move_rotamer_file.split())



import os
import tempfile
import shutil
import pele_platform.Utilities.Helpers.simulation as ad
import pele_platform.Frag.helpers as hp
import pele_platform.Frag.checker as ch
import pele_platform.Frag.parameters.main as mn
import pele_platform.Errors.custom_errors as ce
import frag_pele.main as frag


class FragRunner(mn.FragParameters):

    def __init__(self, args):
        args.system = args.frag_core
        mn.FragParameters.__init__(self, args)

    def run_simulation(self):
        self._set_test_variables()
        self._prepare_control_file()
        self._launch()

    def _launch(self):
        if self.ligands:  # Full ligands as sdf
            fragment_files = self._prepare_input_file(logger=self.logger)
        else:
            fragment_files = None
        self._run()
        
        if self.cleanup and fragment_files:
            self._clean_up(fragment_files)

    def _prepare_control_file(self):
        # Create tmp folder with frag control_file 
        tmp_dir = tempfile.mkdtemp()
        tmp_control_file = os.path.join(tmp_dir, os.path.basename(self.control_file))
        shutil.copy(self.control_file, tmp_control_file)
        adaptive = ad.SimulationBuilder("", tmp_control_file, self)
        # Fill to time because we have flags inside flags
        adaptive.fill_pele_template(self, self.water_object)
        adaptive.fill_pele_template(self, self.water_object)
        self.control_file = tmp_control_file
        return self.control_file

    def _set_test_variables(self):
        if self.test:
            self.gr_steps = 1
            self.frag_steps = 2
            self.frag_eq_steps = 1
            self.temperature = 100000
            self.anm_freq = 0
            self.minimizatoon = 0
            self.sidechain_freq = 0
            self.water_freq = 0
            self.cpus = 4
        else:
            pass

    def _run(self):
        if self.frag_run:
            try:
                frag.main(self.core_process, self.input, self.gr_steps, self.criteria, self.plop_path, self.spython,
                          self.pele_exec, self.control_file, self.license, self.output_folder,
                          self.report_name, "trajectory", self.cluster_folder, self.cpus, self.distcont, self.threshold,
                          self.epsilon, self.condition, self.metricweights,
                          self.nclusters, self.frag_eq_steps, self.frag_restart, self.min_overlap, self.max_overlap,
                          self.chain_core, self.frag_chain, self.frag_steps, self.temperature, self.seed, self.gridres,
                          self.banned, self.limit, self.mae,
                          self.rename, self.threshold_clash, self.steering, self.translation_high, self.rotation_high,
                          self.translation_low, self.rotation_low, self.explorative, self.frag_radius,
                          self.sampling_control, self.pele_data, self.pele_documents,
                          self.only_prepare, self.only_grow, self.no_check, self.debug, srun=self.usesrun)
            except Exception:
                print("Skipped - FragPELE will not run.")

    def _prepare_input_file(self, logger=None):
        from rdkit import Chem
        import rdkit.Chem.rdmolops as rd

        self.input = "input.conf"

        # Check input file
        limit_atoms_ligand = 100
        ch.check_limit_number_atoms(self.ligands, limit_atoms_ligand)
        
        # Get core of the ligand
        mol = Chem.MolFromPDBFile(self.core)
        try:
            ligand_core = rd.SplitMolByPDBResidues(mol)[self.residue]
        except KeyError:
            raise ce.MissResidueFlag("Missing residue flag to specify the ligand core residue name. i.e resname: 'LIG'")
            

        # Get sdf full grown ligands
        ligands_grown = Chem.SDMolSupplier(self.ligands, removeHs=False)
        fragment_files = []
        
        # For each full grown ligand create neutral fragment
        with open(self.input, "w") as fout:
            pass
        for ligand in ligands_grown:
            Chem.AssignAtomChiralTagsFromStructure(ligand)
            try:
                line, fragment = self._create_fragment_from_ligand(ligand, ligand_core, logger=logger)
            except Exception as e:
                try:
                    line, fragment = self._create_fragment_from_ligand(ligand, ligand_core, substructure=False)
                except Exception as e:
                    try:
                        # Try to fix symmetry
                        line, fragment = self._create_fragment_from_ligand(ligand, ligand_core, symmetry=True)
                    except Exception as e:
                            # Try with second substructure search
                        line, fragment = self._create_fragment_from_ligand(ligand, ligand_core, result=1, substructure=False)

            logger.info(f"Ligand {fragment.file} preprocessed")
            fragment_files.append(fragment.file)
            with open(self.input, "a") as fout:
                fout.write(line + "\n")

        return fragment_files

    def _create_fragment_from_ligand(self, ligand, ligand_core, result=0, substructure=True, symmetry=False):
        import rdkit.Chem.AllChem as rp
        from rdkit import Chem

        fragment, old_atoms, hydrogen_core, atom_core, atom_frag, mapping, correct = hp._build_fragment_from_complex(
            self.core, self.residue, ligand, ligand_core, result, substructure, symmetry)

        # temporary override to fix segmentation faults
        filename = "temp.pdb"
        Chem.MolToPDBFile(fragment, filename)
        fragment = Chem.MolFromPDBFile(filename, removeHs=False)
        os.remove(filename)
        
        rp.EmbedMolecule(fragment)
        fragment = hp._retrieve_fragment(
        fragment, old_atoms, atom_core, hydrogen_core, atom_frag, mapping)
        line = fragment.get_inputfile_line()
        fragment.sanitize_file()
        
        if not correct:
            print("Ligand incorrect")
        return line, fragment
        

    def _clean_up(self, fragment_files):

        for file in fragment_files:
            if os.path.isfile(file):
                os.remove(file)



"""
Dictionary with valid flags for platform in input.yalm
"""


VALID_FLAGS_PLATFORM = {"system": "system",
"residue": "resname",
"chain": "chain",
"hbond": "hbond",
"test": "test",
"pele": "pele",
"forcefield": "forcefield",
"verbose": "verbose",
"anm_freq": "anm_freq",
"mpi_params": "mpi_params",
"sidechain_freq": "sidechain_freq",
"min_freq": "min_freq",
"temperature": "temperature",
"sidechain_resolution": "sidechain_res",
"steric_trials": "steric_trials",
"overlap_factor": "overlap_factor",
"steering": "steering",
"solvent": "solvent",
"usesrun": "usesrun",
"spawning": "spawning",
"iterations": "iterations",
"pele_steps": "steps",
"cpus": "cpus",
"density": "density",
"cluster_values": "cluster_values",
"cluster_conditions": "cluster_conditions",
"simulation_type": "simulation_type",
"equilibration": "equilibration",
"eq_steps": "equilibration_steps",
"adaptive_restart": "adaptive_restart",
"input": "global_inputs",
"report_name": "report",
"traj_name": "traj",
"adaptive": "adaptive",
"epsilon": "epsilon",
"bias_column": "bias_column",
"gridres": "gridres",
"core": "core",
"mtor": "maxtorsion",
"n": "n",
"template": "templates",
"ext_temp": "template",
"rotamers": "rotamers",
"mae_lig": "mae_lig",
"skip_prep": "skip_preprocess",
"gaps_ter": "TERs",
"charge_ter": "charge_ters",
"nonstandard": "nonstandard",
"prepwizard": "prepwizard",
"box_center": "box_center",
"box_radius": "box_radius",
"box": "box",
"native": "rmsd_pdb",
"atom_dist": "atom_dist",
"debug": "debug",
"folder": "working_folder",
"output": "output",
"randomize": "randomize",
"full": "global",
"proximityDetection": "proximityDetection",
"poses": "poses",
"precision_glide": "precision_glide",
"msm": "msm",
"precision": "precision",
"clust": "exit_clust",
"restart": "restart",
"lagtime": "lagtime",
"msm_clust": "msm_clust",
"rescoring": "rescoring",
"in_out": "in_out",
"in_out_soft": "in_out_soft",
"exit": "exit",
"exit_value": "exit_value",
"exit_condition": "exit_condition",
"exit_trajnum": "exit_trajnum",
"waters": "waters",
"water_empty_selector": "water_empty_selector", 
"water_freq": "water_freq",
"water_center": "water_center",
"clust_type": "clust_type",
"water_temp": "water_temp",
"water_overlap": "water_overlap",
"water_constr": "water_constr",
"water_trials": "water_trials",
"water_radius": "water_radius",
"out_in": "out_in",
"orthosteric_site": "orthosteric_site",
"final_site": "final_site",
"initial_site" : "initial_site", 
"induced_fit_exhaustive": "induced_fit_exhaustive",
"induced_fit_fast": "induced_fit_fast",
"frag": "frag",
"ca_constr": "ca_constr",
"one_exit": "one_exit",
"box_type": "box_type",
"box_metric": "box_metric",
"time": "time",
"nosasa": "nosasa",
"perc_sasa": "perc_sasa",
"seed": "seed",
"pdb": "pdb",
"log": "log",
"nonrenum": "nonrenum",
"pele_exec": "pele_exec",
"pele_data": "pele_data",
"pele_documents": "pele_documents",
"pca": "pca",
"anm_direction": "anm_direction",
"anm_mix_modes": "anm_mix_modes",
"anm_picking_mode": "anm_picking_mode",
"anm_displacement": "anm_displacement",
"anm_modes_change": "anm_modes_change",
"anm_num_of_modes": "anm_num_of_modes",
"anm_relaxation_constr": "anm_relaxation_constr",
"skip_refinement": "skip_refinement",
"remove_constraints": "remove_constraints",
"pca_traj": "pca_traj",
"perturbation": "perturbation",
"binding_energy": "binding_energy",
"sasa": "sasa",
"parameters": "parameters",
"analyse": "analyse",
"selection_to_perturb": "selection_to_perturb",
"mae": "mae",
"constrain_smiles": "constrain_smiles",
"skip_ligand_prep": "skip_ligand_prep",
"spawning_condition": "spawning_condition",
"external_constraints": "external_constraints",
"only_analysis": "only_analysis",
"overwrite": "overwrite_analysis",
"analysis_nclust": "analysis_nclust",
"te_column": "te_column",
"be_column": "be_column",
"limit_column": "limit_column",
"com": "COMligandConstraint",
"pele_license": "pele_license",
"schrodinger": "schrodinger",
"no_check": "no_check",
"frag_core": "frag_core",
"frag_input": "frag_input",
"frag_ligands": "frag_ligands",
"growing_steps": "growing_steps",
"gpcr_orth": "gpcr_orth",
"frag_steps": "steps_in_gs",
"frag_eq_steps": "sampling_steps",
"protocol": "protocol",
"frag_ai": "frag_ai",
"frag_ai_iterations": "frag_ai_iterations",
"frag_run": "frag_run",
"frag_restart": "frag_restart",
"frag_output_folder": "frag_output_folder", 
"chain_core": "chain_core",
"n_components": "n_components",
"frag_criteria": "frag_criteria",
"frag_cluster_folder": "frag_cluster_folder",
"ppi": "ppi",
"center_of_interface": "center_of_interface", 
"protein": "protein", 
"ligand_pdb": "ligand_pdb",
"allosteric": "allosteric", 
"permissive_metal_constr": "permissive_metal_constr",
"constrain_all_metals": "constrain_all_metals",
"no_metal_constraints": "no_metal_constraints",
"n_waters": "n_waters",
"rna": "rna",
"cleanup": "cleanup",
"polarize_metals": "polarize_metals",
"polarization_factor": "polarization_factor"}



import os
import pele_platform.constants.constants as cs


class FragSimulationParameters:

    def __init__(self, args):
        self.gr_steps = args.growing_steps if args.growing_steps else self.simulation_params.get("growing_steps", 6)
        self.frag_steps = args.frag_steps if args.frag_steps else self.simulation_params.get("steps_in_gs", 3)
        self.frag_eq_steps = args.frag_eq_steps if args.frag_eq_steps else self.simulation_params.get("sampling_steps", 20)
        self.control_file = os.path.join(cs.DIR, "Templates/pele_template.conf")
        self.protocol = args.protocol if args.protocol else self.simulation_params.get("protocol", "")
        self.topology = None if self.pdb else os.path.join("output_pele", "topology.pdb")
        self.frag_restart = "-rst" if args.frag_restart else ""
        self.frag_traj_name = "trajectory"
        self.usesrun = True if self.usesrun == "true" else False



import os
import argparse



class PDBFix():

    def __init__(self, pdb: str, output_folder=":"):
        self.pdb = pdb
        self.lines = self.get_lines()
        self.pdb_out = self.pdb.rsplit(".", 1)[0] + "_proc.pdb"
        self.output_folder = output_folder

    def get_lines(self):
        with open(self.pdb, "r") as f:
            return f.readlines()

    def fix_nucleotides(self):
        new_lines = []
        for line in self.lines:
            try:
                atom_name = line[12:16].strip()
            except IndexError:
                new_lines.append(line)
            if atom_name == "H5'1":
                atom_name = "H5''"
            elif atom_name == "H5'2":
                atom_name = " H5'"
            elif atom_name == "H6 1":
                atom_name = " H61"
            elif atom_name == "H6 2":
                atom_name = " H62"
            elif atom_name == "HN1":
                atom_name = " H1 "
            elif atom_name == "HN3":
                atom_name = " H3 "
            elif atom_name == "H4 1":
                atom_name = " H41"
            elif atom_name == "H4 2":
                atom_name = " H42"
            elif atom_name == "H2 1":
                atom_name = " H21"
            elif atom_name == "H2 2":
                atom_name = " H22"
            else:
                atom_name = line[12:16]
            new_line = line[:12] + atom_name + line[16:]
            new_lines.append(new_line)
        self.lines = new_lines

    def change_K_to_na(self):
        self.lines = [line.replace(" K ", "NA ") if line.startswith("HETATM") else line for line in self.lines]


    def remove_two_first(self):
        residues_to_remove = [line[21:26] for i, line in enumerate(self.lines) if line.startswith("ATOM") and not (self.lines[i+1].startswith("ATOM") and self.lines[i-1].startswith("ATOM"))]
        new_lines = []
        for line in self.lines:
            if line.startswith("ATOM"):
                residue = line[21:26]
            else:
                new_lines.append(line)
                continue
            if residue not in residues_to_remove:
                new_lines.append(line)
        self.lines = new_lines

    def write(self):
        path = os.path.join(self.output_folder, self.pdb_out)
        with open(path, "w") as f:
            f.write("".join(self.lines))
        return path

def fix_rna_pdb(pdb):
    env.logger("Fixing pdb for RNA")
    pdb = PDBFix(pdb)
    pdb.fix_nucleotides()
    pdb.remove_two_first()
    pdb.change_K_to_na()
    output = pdb.write()
    return output

def parse_args(parser):
    parser.add_argument('pdb', type=str, help="Complex pdb with ligand-RNA")
    return parser


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='launch DNA simulation')
    parser = parse_args(parser)
    args = parser.parse_args()
    pdb_fixed = main(args.pdb)
            
               



import pele_platform.Utilities.Helpers.helpers as hp
import pele_platform.constants.constants as cs
import os


class MetricBuilder():


    def __init__(self):
        self.distance = None
        self.rmsd = None

    def distance_to_atom_json(self, pdb, atom_dist):
        distances = []
        for i in range(1, len(atom_dist), 2):
            atom1 = hp.retrieve_atom_info(atom_dist[i-1], pdb)
            atom2 = hp.retrieve_atom_info(atom_dist[i], pdb)
            distances.append(cs.DISTANCE_ATOMS.format(atom1, atom2, i/2))
        self.distance = self._distance_to_json(distances)
        return self.distance


    def _distance_to_json(self, distances):
            return "\n".join(distances)

    def rsmd_to_json(self, pdb_reference, chain_rmsd):
        self.rmsd = cs.NATIVE.format(os.path.abspath(pdb_reference), chain_rmsd)
        return self.rmsd



import glob
import os
import logging
import numpy as np
import sys
import warnings
import pele_platform.Errors.custom_errors as cs
import pele_platform.constants.constants as const
import PPP.global_variables as gv
from Bio.PDB import PDBParser
import pele_platform.Errors.custom_errors as cs

def silentremove(*args, **kwargs):
    for files in args:
        for filename in files:
            try:
                os.remove(filename)
            except OSError:
                pass

def create_dir(base_dir, extension=None):
    """
        Class Method to manage
        directory creation only if that
        ones doesn't exist

        Location:
            base_dir+extension
            or base_dir if extension is None
    """
    if extension:
        path = os.path.join(base_dir, extension)
        if os.path.isdir(path):
            warnings.warn("Directory {} already exists.".format(path), RuntimeWarning)
        else:
            os.makedirs(path)
    else:
        if os.path.isdir(base_dir):
            warnings.warn("Directory {} already exists.".format(base_dir), RuntimeWarning)
        else:
            os.makedirs(base_dir)

class cd:
    """Context manager for changing the current working directory"""
    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


def is_repited(pele_dir):

    original_dir = None
    split_dir = pele_dir.split("_")
    for chunk in split_dir:
        if chunk != "Pele":
            if original_dir:
                original_dir = "{}_{}".format(original_dir, chunk)
            else:
                original_dir = chunk
        else:
            break
    if split_dir[-1].isdigit():
        i = split_dir[-1]
        i = int(i) + 1
    else:
        i = 1
    if os.path.isdir(pele_dir):
                new_pele_dir = "{}_Pele_{}".format(original_dir, i)
                new_pele_dir = is_repited(new_pele_dir)
                return new_pele_dir
    else:
                return pele_dir


def is_last(pele_dir):

    original_dir = None
    split_dir = pele_dir.split("_")
    for chunk in split_dir:
                if chunk != "Pele":
                        if original_dir:
                                original_dir = "{}_{}".format(original_dir, chunk)
                        else:
                                original_dir = chunk
                else:
                        break
    if split_dir[-1].isdigit():
        i = split_dir[-1]
        i = int(i) + 1
    else:
                i = 1

    if os.path.isdir(pele_dir):
            new_pele_dir = "{}_Pele_{}".format(original_dir, i)
            if not os.path.isdir(new_pele_dir):
                return pele_dir
            else:
                            new_pele_dir = is_last(new_pele_dir)
                            return new_pele_dir
    else:
        return pele_dir


def retrieve_atom_info(atom, pdb):
    """
    Parse pdb and return atom name
    chain and residue number
    """
    with open(pdb, "r") as f:
        for line in f:
            try:
                if not isinstance(atom, int) and not atom.isdigit():
                    try:
                        chain, resnum, atomname = atom.split(":")
                    except ValueError:
                        raise ValueError(f"Check atom distance entrance {atom}. Should be like this: 'A:220:OD1'")
                    if line[21].strip() == chain and line[22:26].strip() == resnum and line[12:16].strip() == atomname:
                        atomname = line[12:16]
                        return chain + ":" + resnum + ":" + atomname.replace(" ", "_")
                else:
                    if line[6:11].strip() == str(atom):
                        chain = line[21].strip() 
                        resnum = line[22:26].strip()
                        atomname = line[12:16]
                        return chain + ":" + resnum + ":" + atomname.replace(" ", "_")
            except IndexError:
                pass
        sys.exit(f"Check the atoms {atom} given to calculate the distance metric.")

def retrieve_all_waters(pdb, exclude=False):
    with open(pdb, 'r') as f:
        waters = list(set(["{}:{}".format(line[21:22], line[22:26].strip()) for line in f if line and "HOH" in line]))
    if exclude:
        waters = [water for water in waters if water not in exclude]
    return waters

def retrieve_constraints_for_pele(constraints, pdb):
    CONSTR_ATOM_POINT = '{{ "type": "constrainAtomToPosition", "springConstant": {}, "equilibriumDistance": 0.0, "constrainThisAtom": "{}:{}:{}" }},'
    CONSTR_ATOM_ATOM = '{{"type": "constrainAtomsDistance", "springConstant": {}, "equilibriumDistance": {}, "constrainThisAtom":  "{}:{}:{}", "toThisOtherAtom": "{}:{}:{}"}},'
    final_constraints = []
    for constraint in constraints:
        #Atom to point constraint: 2.2-A:123:2 or 2.2-1986
        if len(constraint.split("-")) == 2:
            spring_constant, atom_info = constraint.split("-")
            chain, residue, atom_name = retrieve_atom_info(atom_info, pdb).split(":")
            constraint = CONSTR_ATOM_POINT.format(spring_constant, chain, residue, atom_name)
        #Atom to atom constraint: 2.2-2.75-A:123:2-B:2:7 or 2.2-2.74-1985-1962
        elif len(constraint.split("-")) == 4:
            spring_constant, eq_distance, atom1_info, atom2_info = constraint.split("-")
            chain1, residue1, atom_name1 = retrieve_atom_info(atom1_info, pdb).split(":")
            chain2, residue2, atom_name2 = retrieve_atom_info(atom2_info, pdb).split(":")
            constraint =  CONSTR_ATOM_ATOM.format(spring_constant, eq_distance, chain1, residue1, atom_name1, chain2, residue2, atom_name2)
        final_constraints.append(constraint)
    return final_constraints
        
def retrieve_box(structure, residue_1, residue_2, weights=[0.5, 0.5]):
    # get center of interface (if PPI)
    coords1 = get_coords_from_residue(structure, residue_1)
    coords2 = get_coords_from_residue(structure, residue_2)
    
    box_center = np.average([coords1, coords2], axis=0, weights=weights)
    box_radius = abs(np.linalg.norm(coords1-coords2))/2 + 4 #Sum 4 to give more space
    return list(box_center), box_radius

def get_coords_from_residue(structure, original_residue):
    parser = PDBParser()
    structure = parser.get_structure('protein', structure)
    chain, res_number, atom_name = original_residue.split(":")
    try:
        res_number = int(res_number)
    except ValueError:
        raise cs.WrongAtomStringFormat(f"The specified atom is wrong '{original_residue}'. \
Should be 'chain:resnumber:atomname'")
    for residue in structure.get_residues():
        if residue.id[1] == res_number:
            for atom in residue.get_atoms():
                if atom.name == atom_name:
                    COI = np.array(list(atom.get_vector()))
                    return COI
    raise cs.WrongAtomSpecified(f"Atom {original_residue} could not be found in structure")


def backup_logger(logger, message):
    if not logger:
        logger = logging.getLogger('logger')
        logger.setLevel(logging.INFO)
        logger.info(message)
    else:
        logger.info(message)

def find_nonstd_residue(pdb):
    with open(pdb, "r") as f:
        resnames = list(set([line[17:20] for line in f \
    if line.startswith("ATOM") and line[17:20] not in gv.supported_aminoacids]))
        return resnames



"""
This file contains de perturbation
parameters of PELE for each type of simulation
"""



OUT_IN = ''' 
             ,
             "parametersChanges" : [
                 { "ifAnyIsTrue": [ "rand >= .5" ],
                     "doThesechanges": { "Perturbation::parameters": { "rotationScalingFactor": 0.1 } },
                     "otherwise": { "Perturbation::parameters": { "rotationScalingFactor": 0.25 } }
                 },
                 { "ifAnyIsTrue": [ "rand1 >= 0.5" ],
                     "doThesechanges": { "Perturbation::parameters": { "translationRange": 3.0, "numberOfTrials" : 10 } },
                     "otherwise": { "Perturbation::parameters": { "translationRange": 1.25, "numberOfTrials" : 10 } }
                 },
                 { "ifAnyIsTrue": [ "sasaLig >= 0.85" ],
                     "doThesechanges": { "Perturbation::parameters": { "translationRange": 5.0, "numberOfTrials" : 20 } },
                     "otherwise": {  }
                 },
                 { "ifAnyIsTrue": [ "sasaLig <= 0.35" ],
                     "doThesechanges": { "Perturbation::parameters": { "rotationScalingFactor": 0.1, "translationRange": 0.75, "numberOfTrials" : 40, "steeringUpdateFrequency": 0} },
                     "otherwise": {  }
                 }
]
'''

GPCR_ORTH = '''
             ,
             "parametersChanges" : [
                 { "ifAnyIsTrue": [ "rand >= .5" ],
                     "doThesechanges": { "Perturbation::parameters": { "rotationScalingFactor": 0.1 } },
                     "otherwise": { "Perturbation::parameters": { "rotationScalingFactor": 0.2 } }
                 },
                 { "ifAnyIsTrue": [ "rand1 >= 0.5" ],
                     "doThesechanges": { "Perturbation::parameters": { "translationRange": 2.0 } },
                     "otherwise": { "Perturbation::parameters": { "translationRange": 1.0 } }
                 },
                 { "ifAnyIsTrue": [ "rand2 >= 0.66" ],
                     "doThesechanges": { "Perturbation::parameters": { "steeringUpdateFrequency": 0, "numberOfTrials" : 10 } },
                     "otherwise": { "Perturbation::parameters": { "steeringUpdateFrequency": 2, "numberOfTrials" : 10 } }
                 },
                 { "ifAnyIsTrue": [ "sasaLig >= 0.85" ],
                     "doThesechanges": { "Perturbation::parameters": { "translationRange": 3.0, "steeringUpdateFrequency": 0,"numberOfTrials" : 40 } },
                     "otherwise": {  }
                 },
                 { "ifAnyIsTrue": [ "sasaLig <= 0.35" ],
                     "doThesechanges": { "Perturbation::parameters": { "rotationScalingFactor": 0.1, "translationRange": 1.0} },
                     "otherwise": {  }
                 }
               ]
'''

IN_OUT = '''
             ,

                 "parametersChanges" : [

                     { "ifAnyIsTrue": [ "rand >= .5" ],

                         "doThesechanges": { "Perturbation::parameters": { "rotationScalingFactor": 0.05 } },

                         "otherwise": { "Perturbation::parameters": { "rotationScalingFactor": 0.15 } }

                     },

                     {

                       "ifAnyIsTrue": [ "rand1 >= 0.40" ],

                         "doThesechanges": { "Perturbation::parameters": { "translationRange": 2.0 } },

                         "otherwise": { "Perturbation::parameters": { "translationRange": 2.0 } }

                    },
                    {
                      "ifAnyIsTrue": [ "rand3 >= 0.10" ],
                        "doThesechanges": { "Perturbation::parameters": { "steeringUpdateFrequency": 1, "numberOfTrials" : 10 } },
                        "otherwise": { "Perturbation::parameters": { "steeringUpdateFrequency": 0, "numberOfTrials" : 25 } }
                    },
                    {
                      "ifAnyIsTrue": [ "sasaLig <= 0.15" ],
                         "doThesechanges": { "Pele::parameters": { "perturbationCOMConstraintConstant" : 0.25 }, "Perturbation::parameters": { "translationRange": 1.0 }},
                         "otherwise": { "Pele::parameters": { "perturbationCOMConstraintConstant" : 1.0 } }
                    }, 
                    {
                      "ifAnyIsTrue": [ "sasaLig >= 0.75" ],
                         "doThesechanges": { "Pele::parameters": { "perturbationCOMConstraintConstant" : 10.0 }, "Perturbation::parameters": { "steeringUpdateFrequency": 1, "numberOfTrials" : 4 }},
                         "otherwise": { }
                    } 

                  ]

'''

RESCORING = '''
             ,
             "parametersChanges" : [
             
                  { "ifAnyIsTrue": [ "rand >= .5" ],
             
                      "doThesechanges": { "Perturbation::parameters": { "rotationScalingFactor": 0.1 } },
             
                      "otherwise": { "Perturbation::parameters": { "rotationScalingFactor": 0.05 } }
             
                  },
             
                  { "ifAnyIsTrue": [ "rand1 >= 0.5" ],
             
                      "doThesechanges": { "Perturbation::parameters": { "translationRange": 0.25} },
             
                      "otherwise": { "Perturbation::parameters": { "translationRange": 0.5} }
             
                  },
             
                  {  "ifAnyIsTrue": [ "rand2 >= 0.5" ],
             
                         "doThesechanges": {  "Perturbation::parameters": { "steeringUpdateFrequency": 0, "numberOfTrials": 25 } },
             
                         "otherwise": {  "Perturbation::parameters": { "steeringUpdateFrequency": 0 , "numberOfTrials": 25  }}
             
                 }
             
             ]
'''



INDUCED_FIT = '''
             ,
"parametersChanges" : [
     { "ifAnyIsTrue": [ "rand >= .5" ],
         "doThesechanges": { "Perturbation::parameters": { "rotationScalingFactor": 0.1 } },
         "otherwise": { "Perturbation::parameters": { "rotationScalingFactor": 0.25 } }
     },
     { "ifAnyIsTrue": [ "rand1 >= 0.5" ],
         "doThesechanges": { "Perturbation::parameters": { "translationRange": 0.5} },
         "otherwise": { "Perturbation::parameters": { "translationRange": 1.0} }
     },
     {  "ifAnyIsTrue": [ "rand2 >= 0.5" ],
            "doThesechanges": {  "Perturbation::parameters": { "steeringUpdateFrequency": 0, "numberOfTrials": 30 } },
            "otherwise": {  "Perturbation::parameters": { "steeringUpdateFrequency": 1 , "numberOfTrials": 10  }}
    }
]
'''


GLOBAL = '''
             ,
             "parametersChanges" : [
                 { "ifAnyIsTrue": [ "rand >= .5" ],
                     "doThesechanges": { "Perturbation::parameters": { "rotationScalingFactor": 0.1 } },
                     "otherwise": { "Perturbation::parameters": { "rotationScalingFactor": 0.25 } }
                 },
                 { "ifAnyIsTrue": [ "rand1 >= 0.5" ],
                     "doThesechanges": { "Perturbation::parameters": { "translationRange": 3.0, "numberOfTrials" : 10 } },
                     "otherwise": { "Perturbation::parameters": { "translationRange": 1.25, "numberOfTrials" : 10 } }
                 },
                 { "ifAnyIsTrue": [ "sasaLig >= 0.85" ],
                     "doThesechanges": { "Perturbation::parameters": { "translationRange": 5.0, "numberOfTrials" : 20 } },
                     "otherwise": {  }
                 },
                 { "ifAnyIsTrue": [ "sasaLig <= 0.35" ],
                     "doThesechanges": { "Perturbation::parameters": { "rotationScalingFactor": 0.1, "translationRange": 0.75, "numberOfTrials" : 40, "steeringUpdateFrequency": 0} },
                     "otherwise": {  }
                 }
              ]
'''

WATER_PARAMS = '''
                     ,{
                         "ifAnyIsTrue": [ "rand3 <= 0.4" ],
                         "doThesechanges": { "WaterPerturbation::parameters": { "translationRange": 4.0 } },
                         "otherwise": { "WaterPerturbation::parameters": { "translationRange": 3.0} }
                     },
                     {
                         "ifAnyIsTrue": [ "rand3 >= 0.85" ],
                         "doThesechanges": { "WaterPerturbation::parameters": { "translationRange": 2.0 } }
                     }]
'''


FRAG = '''
,
                 "parametersChanges" : [
                     { "ifAnyIsTrue": [ "rand >= 0.5" ],
                         "doThesechanges": { "Perturbation::parameters": { "rotationScalingFactor": $ROTATION_HIGH } },
                         "otherwise": { "Perturbation::parameters": { "rotationScalingFactor": $ROTATION_LOW } }
                     },
                     { "ifAnyIsTrue": [ "rand1 <= 0.5 " ],
                         "doThesechanges": { "Perturbation::parameters": { "translationRange": $TRANSLATION_LOW } },
                         "otherwise": { "Perturbation::parameters": { "translationRange": $TRANSLATION_HIGH } }
                     }
                  ]
'''



import random
import os
import glob
import pele_platform.constants.constants as cs
from pele_platform.Utilities.Parameters.SimulationParams.MSMParams import msm_params
from pele_platform.Utilities.Parameters.SimulationParams.GlideParams import glide_params
from pele_platform.Utilities.Parameters.SimulationParams.BiasParams import bias_params
from pele_platform.Utilities.Parameters.SimulationParams.InOutParams import inout_params
from pele_platform.Utilities.Parameters.SimulationParams.PCA import pca
from pele_platform.Utilities.Parameters.SimulationParams.Allosteric import allosteric
from pele_platform.Utilities.Parameters.SimulationParams.PPI import ppi
#from pele_platform.Utilities.Parameters.SimulationParams.RNA import rna
import pele_platform.Utilities.Helpers.helpers as hp

LOGFILE = '"simulationLogPath" : "$OUTPUT_PATH/logFile.txt",'


class SimulationParams(msm_params.MSMParams, glide_params.GlideParams, bias_params.BiasParams, 
    inout_params.InOutParams, pca.PCAParams, allosteric.AllostericParams, ppi.PPIParams): # rna.RNAParams


    def __init__(self, args):
        self.simulation_type(args)
        self.main_pele_params(args)
        self.main_adaptive_params(args)
        self.optative_params(args)
        self.system_preparation_params(args)
        self.ligand_params(args)
        self.anm_params(args)
        self.water_params(args)
        self.box_params(args)
        self.output_params(args)
        self.analysis_params(args)

        #Create all simulation types (could be more efficient --> chnage in future) 
        super().generate_msm_params(args)
        super().generate_glide_params(args)
        super().generate_bias_params(args)
        super().generate_inout_params(args)
        super().generate_pca_params(args)
        super().generate_allosteric_params(args)
        super().generate_ppi_params(args)
        #rna.RNAParams.__init__(self, args)


    def simulation_type(self, args):
        self.adaptive = True if args.package in ["allosteric", "adaptive", "PPI"]  else None
        self.frag_pele = True if args.package == "frag" else None
        # Trick to let frag handle control fodler parameters --> Improve
        self.complexes = "$PDB" if self.frag_pele else "$COMPLEXES"
        self.frag_pele_steps = "$STEPS" if self.frag_pele else "$PELE_STEPS"
        self.output_path = "$RESULTS_PATH" if self.frag_pele else "$OUTPUT_PATH"

    def main_pele_params(self,args):
        if "*" in args.system:
            self.system = glob.glob(args.system)[0]
            args.input = glob.glob(args.system)
        else:
            self.system = args.system
        self.residue = args.residue
        self.chain = args.chain
        if self.adaptive:
            assert self.system and self.residue and self.chain, "User must define input, residue and chain"
        self.debug = args.debug if args.debug else False
        self.pele_steps = args.pele_steps if args.pele_steps else self.simulation_params.get("pele_steps", 8)
        self.logfile =  LOGFILE if args.log else ""
        self.license = args.pele_license if args.pele_license else os.path.join(cs.PELE, "licenses")
        self.anm_freq = args.anm_freq if args.anm_freq is not None else self.simulation_params.get("anm_freq", 4)
        self.sidechain_freq = args.sidechain_freq if args.sidechain_freq is not None else self.simulation_params.get("sidechain_freq", 2)
        self.min_freq = args.min_freq if args.min_freq is not None else self.simulation_params.get("min_freq", 1)
        self.water_freq = args.water_freq if args.water_freq is not None else self.simulation_params.get("water_freq", 1)
        self.temperature = args.temperature if args.temperature else self.simulation_params.get("temperature", 1500)
        self.sidechain_resolution = args.sidechain_resolution if args.sidechain_resolution else self.simulation_params.get("sidechain_resolution", 30)
        self.proximityDetection = "false" if args.proximityDetection is False else self.simulation_params.get("proximityDetection", "true")
        self.steric_trials = args.steric_trials if args.steric_trials else self.simulation_params.get("steric_trials", 250)
        self.ca_constr = args.ca_constr if args.ca_constr is not None else self.simulation_params.get("ca_constr", 0.5)
        self.overlap_factor = args.overlap_factor if args.overlap_factor else self.simulation_params.get("overlap_factor", 0.65)
        self.steering= args.steering if args.steering else self.simulation_params.get("steering", 0)
        self.perturbation = "" if args.perturbation is False else self.simulation_params.get("perturbation", cs.PERTURBATION)
        self.perturbation_params(args)
        self.com = args.com if args.com else self.simulation_params.get("COMligandConstraint", 0)

    def anm_params(self, args):
        self.anm_displacement = args.anm_displacement if args.anm_displacement else self.simulation_params.get("anm_displacement", 0.75)
        self.anm_modes_change = args.anm_modes_change if args.anm_modes_change else self.simulation_params.get("anm_modes_change", 4)
        self.anm_direction = args.anm_direction if args.anm_direction else self.simulation_params.get("anm_direction", "random")
        self.anm_mix_modes = args.anm_mix_modes if args.anm_mix_modes else self.simulation_params.get("anm_mix_modes", "mixMainModeWithOthersModes")
        self.anm_picking_mode = args.anm_picking_mode if args.anm_picking_mode else self.simulation_params.get("anm_picking_mode", "RANDOM_MODE")
        self.anm_num_of_modes = args.anm_num_of_modes  if args.anm_num_of_modes else self.simulation_params.get("anm_num_of_modes", 6)
        self.anm_relaxation_constr = args.anm_relaxation_constr if args.anm_relaxation_constr else self.simulation_params.get("anm_relaxation_constr", 0.5)
        self.remove_constraints = args.remove_constraints if args.remove_constraints is not None else self.simulation_params.get("remove_constraints", False)

    def perturbation_params(self, args):
        if self.perturbation:
            self.selection_to_perturb = args.selection_to_perturb if args.selection_to_perturb else self.simulation_params.get("selection_to_perturb", cs.SELECTION_TO_PERTURB)
            self.parameters = args.parameters  if args.parameters else self.simulation_params.get("params", True)
            self.ligand = cs.LIGAND if self.perturbation else ""
            self.binding_energy = args.binding_energy if args.binding_energy else self.simulation_params.get("binding_energy", cs.BE)
            self.sasa = args.sasa if args.sasa else self.simulation_params.get("sasa", cs.SASA)
        else:
            self.selection_to_perturb = ""
            self.parameters = ""
            self.ligand = ""
            self.binding_energy = ""
            self.sasa = ""


    def main_adaptive_params(self, args):
        self.spawning = args.spawning if args.spawning else self.simulation_params.get("spawning_type", "independent")
        self.spawning_condition = args.spawning_condition if args.spawning_condition else self.simulation_params.get("spawning_condition", None)
        self.spawning_condition = '"condition": "{}",'.format(self.spawning_condition) if self.spawning_condition else ""
        self.density = args.density if args.density else self.simulation_params.get("density", "null")
        self.simulation_type = args.simulation_type if args.simulation_type else self.simulation_params.get("simulation_type", "pele")
        iterations = 1 if self.spawning == "independent" else 30
        self.iterations = args.iterations if args.iterations else self.simulation_params.get("iterations", iterations)
        self.clust_type = args.clust_type if args.clust_type else self.simulation_params.get("clust_type", "rmsd")
        self.cluster_values = args.cluster_values if args.cluster_values else self.simulation_params.get("cluster_values", "[1.75, 2.5, 4, 6]")
        self.cluster_conditions = args.cluster_conditions if args.cluster_conditions else self.simulation_params.get("cluster_conditions", "[1, 0.6, 0.4, 0.0]")
        self.seed = args.seed if args.seed else random.randrange(1, 70000)
        self.templates = os.path.abspath(os.path.join(os.path.dirname(os.path.dirname(__file__)), "PeleTemplates"))
        self.usesrun = "true" if args.usesrun else "false"
        self.mpi_params = f'"mpiParameters": "{args.mpi_params}",' if args.mpi_params else ""

    def optative_params(self, args):
        self.input = args.input
        self.forcefield = args.forcefield if args.forcefield else self.simulation_params.get("forcefield", "OPLS2005")
        self.solvent = args.solvent if args.solvent else self.simulation_params.get("solvent", "VDGBNP") 
        self.verbose = "true" if args.verbose else self.simulation_params.get("verbose", "false")
        self.cpus = args.cpus = args.cpus if args.cpus else self.simulation_params.get("cpus", 60)
        self.restart = args.restart if args.restart else self.simulation_params.get("restart", "all")
        self.test = args.test 
        #+1 to avoid being 0 
        self.equil_steps = int(args.eq_steps/self.cpus) + 1 if args.eq_steps else self.simulation_params.get("equilibration_steps", 1)
        self.equilibration = "true" if args.equilibration else self.simulation_params.get("equilibration", "false")
        self.adaptive_restart = args.adaptive_restart
        self.poses = args.poses if args.poses else self.simulation_params.get("poses", self.cpus-1)
        self.pele_exec = args.pele_exec if args.pele_exec else os.path.join(cs.PELE, "bin/Pele_mpi")
        self.pele_data = args.pele_data if args.pele_data else os.path.join(cs.PELE, "Data")
        self.pele_documents = args.pele_documents if args.pele_documents else os.path.join(cs.PELE, "Documents")
        self.polarize_metals = args.polarize_metals if args.polarize_metals else False
        self.polarization_factor = args.polarization_factor if args.polarization_factor else 2.0

    def system_preparation_params(self, args):
        self.skip_prep = args.skip_prep if args.skip_prep else self.simulation_params.get("skip_prep", False)
        self.nonstandard = args.nonstandard if args.nonstandard else self.simulation_params.get("nonstandard", [])
        self.constraints = None
        self.external_constraints = hp.retrieve_constraints_for_pele(args.external_constraints, self.system) if args.external_constraints else []
        self.permissive_metal_constr = args.permissive_metal_constr if args.permissive_metal_constr else []
        self.constrain_smiles = args.constrain_smiles if args.constrain_smiles else self.simulation_params.get("constrain_smiles", None)
        self.no_ppp = args.no_ppp if args.no_ppp else self.simulation_params.get("no_ppp", False)

    def ligand_params(self, args):
        self.spython = os.path.join(cs.SCHRODINGER, "utilities/python")
        if not os.path.exists(self.spython):
            self.spython = os.path.join(cs.SCHRODINGER, "run")
        self.mae_lig = args.mae_lig if args.mae_lig else self.simulation_params.get("mae_lig", None)
        self.external_template = args.template if args.template else self.simulation_params.get("template", [])
        self.external_rotamers = args.rotamers if args.rotamers else self.simulation_params.get("rotamers", [])
        self.skip_ligand_prep = args.skip_ligand_prep if args.skip_ligand_prep else self.simulation_params.get("args.skip_ligand_prep", [])
        self.core = args.core
        self.n = args.n
        self.mtor = args.mtor
        self.forcefield = args.forcefield
        self.mae_lig = args.mae_lig
        self.lig = self.mae_lig if self.mae_lig else "{}.mae".format(self.residue)
        self.gridres = args.gridres

    def water_params(self, args):
        self.water_temp = args.water_temp if args.water_temp else self.simulation_params.get("water_temp", 5000)
        self.water_overlap = args.water_overlap if args.water_overlap else self.simulation_params.get("water_overlap", 0.78)
        self.water_constr = args.water_constr if args.water_constr else self.simulation_params.get("water_constr", 0)
        self.water_trials = args.water_trials if args.water_trials  else self.simulation_params.get("water_trials", 10000)

       
        self.allow_empty_selectors = '"allowEmptyWaterSelectors": true,' if args.water_empty_selector else ""
        self.water_arg = hp.retrieve_all_waters(self.system) if args.waters == "all_waters" else args.waters  # IDs of waters
        self.n_waters = args.n_waters if args.n_waters  else self.simulation_params.get("n_waters", 0)
        self.waters = args.waters if args.waters else self.simulation_params.get("waters", [])
        self.water_radius = args.water_radius if args.water_radius else 6
        self.water_center = None
        self.water = ""
        self.water_energy = None


    def box_params(self, args):
        self.box_radius = args.box_radius if args.box_radius else self.simulation_params.get("box_radius", None)
        if args.box_center:
            if ":" in args.box_center:
                self.box_center = [str(coord) for coord in hp.get_coords_from_residue(self.system, args.box_center)]
                self.box_center = "[" + ", ".join(self.box_center) + "]"
            else:
                self.box_center = [str(x) for x in args.box_center] 
                self.box_center = "["+ ",".join([str(coord) for coord in self.box_center]) + "]"
        else:
            self.box_center = self.simulation_params.get("box_center", None)

    def output_params(self, args):
        self.folder = args.folder
        self.output = args.output if args.output is not None else "output"
        self.report_name = args.report_name if args.report_name else "report"
        self.traj_name = args.traj_name if args.traj_name else "trajectory.pdb"
        self.xtc = self.traj_name.endswith(".xtc")
        self.pdb = self.traj_name.endswith(".pdb")

    def analysis_params(self, args):
        self.analyse = args.analyse if args.analyse else True
        self.mae = args.mae if args.mae else False
        self.only_analysis = args.only_analysis
        self.analysis_nclust = args.analysis_nclust
        self.overwrite = args.overwrite
        self.be_column = args.be_column
        self.te_column = args.te_column
        self.limit_column = args.limit_column



import os
import pele_platform.Utilities.Helpers.plop_launcher as plop
import pele_platform.Utilities.Helpers.system_prep as sp

def create_template(env, residue):
    output_pdb = os.path.join(env.pele_dir, "miss_residue.pdb")
    syst = sp.SystemBuilder.build_system(env.system_fix, None, residue, env.pele_dir, output=output_pdb)
    plop.parametrize_miss_residues(env, residue, syst.lig)



class Atom:

    def __init__(self, molecule, idx):
        self.molecule = molecule
        self.idx = idx

    def get_name(self):
        self.name = self.molecule.GetAtomWithIdx(self.idx).GetMonomerInfo().GetName().strip()
        return self.name



from dataclasses import dataclass
import pele_platform.Adaptive.simulation as si
import pele_platform.Utilities.Helpers.helpers as hp
import pele_platform.Utilities.Parameters.pele_env as pv


@dataclass
class GpcrLauncher:

    args: pv.EnviroBuilder 

    def run_gpcr_simulation(self) -> pv.EnviroBuilder:
        # Set parameters for GPCR and launch simulation
        self._set_parameters()
        simulation_parameters = si.run_adaptive(self.args)
        return simulation_parameters
        

    def _set_parameters(self) -> None:
        # Set box and initial ligand position
        self.orthosteric_site = self.args.orthosteric_site
        self.initial_site = self.args.initial_site
        self.args.center_of_interface = self.initial_site
        box_center, box_radius = hp.retrieve_box(
    self.args.system, self.initial_site, self.orthosteric_site,
    weights=[0.35, 0.65])
        self.args.box_center = self.args.box_center if self.args.box_center else box_center
        self.args.box_radius = self.args.box_radius if self.args.box_radius else box_radius
        self.args.randomize = True



import os
from string import Template


class TemplateBuilder(object):

    def __init__(self, file, keywords):

        self.file = file
        self.keywords = {k: v for k, v in keywords.items() if v is not None}

        self.fill_in()

    def fill_in(self):
        """
        Fill the control file in
        """
        with open(os.path.join(self.file), 'r') as infile:
            confile_data = infile.read()

        confile_template = Template(confile_data)

        confile_text = confile_template.safe_substitute(self.keywords)

        with open(os.path.join(self.file), 'w') as outfile:
            outfile.write(confile_text)



from dataclasses import dataclass
import os
import pele_platform.constants.constants as cs
from pele_platform.Checker import executable as ex
from pele_platform.Checker import environment_variables as en
import pele_platform.Utilities.Helpers.yaml_parser as yp


EXECUTABLES_TO_CHECK = ["mpirun",]


@dataclass
class Checker():

    def check_variables(self, args: yp.YamlParser) -> None:
        for env_variable in self._generate_env_variables(args):
            env_variable.is_valid()
        for executable in self._generate_executables():
            executable.is_in_path()

    def _generate_env_variables(self, args: yp.YamlParser) -> list:
        self.env_variables = [
        en.EnvVariable("pele_data", args.pele_data, os.path.join(cs.PELE, "Data"), "--pele_data /path/to/data/folder/", "export PELE=/path/to/PELE-1.X/"),
        en.EnvVariable("pele_documents", args.pele_documents, os.path.join(cs.PELE, "Documents"), "--pele_documents /path/to/documents/folder", "export PELE=/path/to/PELE-1.X/"),
        en.EnvVariable("pele_exec", args.pele_exec, os.path.join(cs.PELE, "bin/Pele_mpi"), "--pele_exec /path/to/PELE_exec", "export PELE=/path/to/PELE-1.X/"),
        en.EnvVariable("pele_license", args.pele_license, os.path.join(cs.PELE, "licenses"), "--pele_license /path/to/licenses", "export PELE=/path/to/PELE-1.X/"),
        en.EnvVariable("schrodinger", args.schrodinger, cs.SCHRODINGER, "--schrodinger /path/to/schrodinger-20XX/", "export SCHRODINGER=/path/to/schrodinger-20XX/")
        ]
        return self.env_variables


    def _generate_executables(self) -> list:
        self.executables = [ex.Executable(executable) for executable in EXECUTABLES_TO_CHECK]
        return self.executables


    
def check_executable_and_env_variables(args: yp.YamlParser):
    """
    Check all external requirements are there
    before starting the simulation.
    1) Check env variables
    2) Check executables 
    """
    checker = Checker()
    checker.check_variables(args)



from dataclasses import dataclass
import logging
import pele_platform.Utilities.Helpers.center_of_mass as cm
import pele_platform.constants.constants as cs

@dataclass
class BoxSetter:
    '''
    Base class to generate the
    exploration box for the simulation
    '''
    box_center: list
    box_radius: float
    ligand_ref: str=""
    logger: logging.Logger=None

    def generate_json(self) -> str:
        # From a box center and box radius generates 
        # json string for PELE control file
        self.logger.info("Generating exploration box")
        self._set_box_center()
        self.logger.info(f"Box with center {self.box_center} \
and radius {self.box_radius} generated\n\n")
        return self.box_json

    def _set_box_center(self):
        # If not user's box center
        # set center of mass of the ligand
        if not self.box_center:
            self.box_center = cm.center_of_mass(self.ligand_ref)

    @property
    def box_json(self):
        # generate the json format for the box
        return cs.BOX.format(self.box_radius, self.box_center) if  self.box_radius else ""




import random
import string
import pele_platform.Frag.fragments as fr
import pele_platform.Frag.atoms as at
import pele_platform.Frag.checker as ch
import pele_platform.Errors.custom_errors as ce



def _search_core_fragment_linker(ligand, ligand_core, result=0, check_symmetry=False):
    """ Given mol1 and mol2 return the linker atoms"""
    substructure_results = ligand.GetSubstructMatches(ligand_core)

    try:
        print("Running substructure search on {}...".format(ligand.GetProp('_Name')))
        core_atoms = substructure_results[result]
        print("RESULT SUBSTRUCTURE", core_atoms)
    except IndexError:
        raise IndexError(
            "Make sure core from pdb and full fragment are the same. Make sure that both core and fragment have "
            "correctly defined aromatic bonds. Also, all fragments must have different molecule name.")

    # Sometimes substructure search messes up symmetry. Check that!
    if check_symmetry:
        core_atoms = ch.check_substructure_match(ligand, ligand_core, core_atoms)
    for atom in ligand.GetAtoms():
        if atom.GetIdx() in core_atoms or atom.GetAtomicNum() == 1:
            continue
        else:
            atoms_bonded = atom.GetNeighbors()
            for neighbour in atoms_bonded:
                if neighbour.GetIdx() in core_atoms:
                    return core_atoms.index(neighbour.GetIdx()), core_atoms, atom.GetIdx()


def _build_fragment_from_complex(complex, residue, ligand, ligand_core, result=0, substructure=True, simmetry=False):
    from rdkit import Chem
    import rdkit.Chem.rdmolops as rd
    import rdkit.Chem.rdchem as rc

    # Retrieve atom core linking fragment
    try:
        atom_core_idx, atoms_core, atom_fragment = _search_core_fragment_linker(ligand, ligand_core, result, simmetry)
        print("ATOM OF FRAGMENT ATTACHED TO CORE:", atom_fragment)
        print("ATOM OF CORE ATTACHED TO FRAGMENT:", atom_core_idx)
    except TypeError:
        raise ce.SameMolecule("Core and ligand are the exact same molecule. Check your inputs.")
    atom_core = at.Atom(ligand_core, atom_core_idx)
    mol = Chem.MolFromPDBFile(complex, removeHs=False)

    # Retrieve hydrogen core attach to linking atom
    original = rd.SplitMolByPDBResidues(mol)[residue]
    hydrogen_core_idx = \
        [atom.GetIdx() for atom in original.GetAtomWithIdx(atom_core_idx).GetNeighbors() if atom.GetAtomicNum() == 1][0]
    hydrogen_core = at.Atom(original, hydrogen_core_idx)


    # Delete core for full ligand with substructure and if it fails manually
    if substructure:
        fragment = rd.DeleteSubstructs(ligand, ligand_core)
        new_mol = rc.EditableMol(fragment)
        for atom in reversed(fragment.GetAtoms()):
            neighbours = atom.GetNeighbors()
            if len(neighbours) == 0 and atom.GetAtomicNum() == 1:
                new_mol.RemoveAtom(atom.GetIdx())
    else:
        new_mol = rc.EditableMol(ligand)

        for atom in reversed(atoms_core):
            new_mol.RemoveAtom(atom)


        for atom in reversed(new_mol.GetMol().GetAtoms()):
            neighbours = atom.GetNeighbors()
            if len(neighbours) == 0 and atom.GetAtomicNum() == 1:
                new_mol.RemoveAtom(atom.GetIdx())
    print("FRAGMENT ATOMS", [atom.GetIdx() for atom in new_mol.GetMol().GetAtoms()])

    # Add missing hydrogen to full ligand and create pdb differently depending on the previous step
    fragment = new_mol.GetMol()
    old_atoms = [atom.GetIdx() for atom in fragment.GetAtoms() if atom.GetAtomicNum() != 1]
    new_atoms = [atom.GetIdx() for atom in ligand.GetAtoms() if
                 atom.GetIdx() not in atoms_core and atom.GetAtomicNum() != 1]
    print("old_atoms", old_atoms, "new_atoms", new_atoms)
    mapping = {new_atom: old_atom for new_atom, old_atom in zip(new_atoms, old_atoms)}
    atom_fragment_mapped = mapping[atom_fragment]

    fragment = Chem.AddHs(fragment, False, True)
    correct = _check_fragment(fragment, ligand, mapping, atom_fragment, atom_fragment_mapped, ligand_core)
    Chem.MolToPDBFile(fragment, "int0.pdb")
    assert len(old_atoms) == len(new_atoms)
    return fragment, old_atoms, hydrogen_core, atom_core, atom_fragment, mapping, correct


def random_string(string_length=8):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(string_length))


def _retrieve_fragment(fragment, old_atoms, atom_core, hydrogen_core, atom_fragment, mapping):
    from rdkit import Chem

    # Get new added hydrogen
    try:
        added_hydrogen_idx = [atom.GetIdx() for atom in fragment.GetAtoms() if atom.GetIdx() not in old_atoms][0]
        no_hydrogens = False
        # Get fragment atom attached to newly added hydrogen
        atom_fragment_idx = fragment.GetAtomWithIdx(added_hydrogen_idx).GetNeighbors()[0].GetIdx()
    except IndexError:
        logger.info("Hydrogen detection failed won't have into account steriochemistry")
        added_hydrogen_idx = 0
        no_hydrogens = True
        atom_fragment_idx = mapping[atom_fragment]
        print("Final atom fragment", atom_fragment_idx)

    # Save fragment
    string = random_string(8)
    fragment_filename = fragment.GetProp("_Name").replace(" ", "_") + string + ".pdb"

    Chem.MolToPDBFile(fragment, fragment_filename)
    fragment = Chem.MolFromPDBFile(fragment_filename, removeHs=False)
    added_hydrogen = at.Atom(fragment, added_hydrogen_idx)
    atom_fragment_attach_to_hydrogen = at.Atom(fragment, atom_fragment_idx)

    # Build fragment object
    fragment = fr.Fragment(fragment_filename, atom_fragment_attach_to_hydrogen,
                           added_hydrogen, atom_core, hydrogen_core, no_hydrogens)
    return fragment


def _check_fragment(fragment, ligand, mapping, atom_fragment, atom_fragment_mapped, ligand_core):
    from rdkit import Chem

    frag_neighbours = fragment.GetAtomWithIdx(atom_fragment_mapped).GetNeighbors()
    lig_neighbours = ligand.GetAtomWithIdx(atom_fragment).GetNeighbors()

    # Check, if atom_fragments has correct number of hydrogens by comparing to the original ligand
    if len(frag_neighbours) < len(lig_neighbours):
        to_add = len(lig_neighbours) - len(frag_neighbours)
        fragment.GetAtomWithIdx(atom_fragment_mapped).SetNumExplicitHs(to_add)

    # Add hydrogens
    fragment = Chem.AddHs(fragment, False, True)

    # Compare chirality
    lig_chir = Chem.FindMolChiralCenters(ligand, force=True, includeUnassigned=True)
    frag_chir = Chem.FindMolChiralCenters(fragment, force=True, includeUnassigned=True)

    # Check fragment size
    ligand_len = ligand.GetNumHeavyAtoms()
    frag_len = fragment.GetNumHeavyAtoms()
    core_len = ligand_core.GetNumHeavyAtoms()

    if ligand_len - core_len == frag_len:
        if lig_chir and frag_chir:
            for lg, fg in zip(lig_chir, frag_chir):
                # If chirality different or atom number different --> ? as incorrect
                if lg[1] != fg[1] or mapping[lg[0]] != fg[0]:
                    if lg[1] != "?" and fg[1] != "?":
                        correct = False
                        break
                    else:
                        correct = True
                else:
                    correct = True
        else:
            correct = True
    else:
        correct = False

    return correct



import glob
import os
import pele_platform.Utilities.Helpers.Metals.metal_constraints as mc

def change_metal_charges(template_dir, forcefield, polarisation_factor, protein_file):

    # get names of all metals in the structure
    metals_in_structure, _ = mc.find_metals(protein_file)
    metals = list(set([atom.name for atom, _, _ in metals_in_structure]))

    # get Data and LocalData directories
    template_paths = []
    local_template_dir = template_dir
    global_template_dir = os.path.join(os.environ["PELE"], "Data")

    template_paths = find_metal_templates(metals, forcefield, global_template_dir)
    change_polarisation(template_paths, local_template_dir, polarisation_factor)


def find_metal_templates(metals, forcefield, global_template_dir):

    # find all metal template files in PELE Data
    template_paths = []
    
    for metal in metals:
        metal_template = "{}z".format(metal.lower())
        metal_template_file = glob.glob(os.path.join(global_template_dir, "Templates/{}/HeteroAtoms".format(forcefield), metal_template))
        if metal_template_file:
            template_paths.append(metal_template_file[0])
    
    return template_paths


def change_polarisation(template_paths, local_template_dir, polarisation_factor):

    # get the line with metal charge and divide the value by two
    for path in template_paths:
        with open(path, "r") as file_in:
            lines = file_in.readlines()

            for i in range(len(lines)):
                if "NBON" in lines[i]:
                    charge = lines[i+1][26:33]
                    charge_after = "{:7.6f}".format(float(charge)/polarisation_factor)
                    lines[i+1] = lines[i+1][:25] + charge_after + lines[i+1][34:]

        # create a path in DataLocal
        copied_path = os.path.join(local_template_dir, os.path.basename(path))

        # write to DataLocal
        with open(copied_path, "w+") as file_out:
            for line in lines:
                file_out.write(line)




class Fragment:

    def __init__(self, file, atom_fragment, hydrogen_fragment, atom_core, hydrogen_core, no_hydrogen):
        from rdkit import Chem
        self.file = file
        self.atom_fragment = atom_fragment
        self.hydrogen_fragment = hydrogen_fragment
        self.atom_core = atom_core
        self.hydrogen_core = hydrogen_core
        self.mol = Chem.MolFromPDBFile(self.file, removeHs=False)
        self.no_hydrogen = no_hydrogen

    def get_inputfile_line(self):
        if self.no_hydrogen:
            self.line = "{} {} {}".format(self.file, 
                self.atom_core.get_name(),
                self.atom_fragment.get_name())
        else:
            self.line = "{} {}-{} {}-{}".format(self.file, 
                self.atom_core.get_name(), self.hydrogen_core.get_name(),
                self.atom_fragment.get_name(), self.hydrogen_fragment.get_name())
        return self.line

    def sanitize_file(self):
        with open(self.file, "r") as fin:
            lines = fin.readlines()
        with open(self.file, "w") as fout:
            new_lines = [ line[:17] + "FRG L" + line[22:] if line.startswith("HETATM") else line for line in lines]
            fout.write("".join(new_lines))



import os

############################
# PRIVATE CONSTANTS
#############################

SCHRODINGER = os.environ.get("SCHRODINGER", "")
PELE = os.environ.get("PELE", "")

# DEFAULTS
# --------

COMPLEX = "complex.pdb"
RESULTS = "results"
LIG_RES = "LIG"
LIG_CHAIN = "Z"
FORCEFIELD = "OPLS2005"
PELE_CONFILE = "pele.conf"
CPUS = 140
RESTART = "true"
CLUSTERS = 40
PLATFORM_RESTART = "all"
EQ_STEPS = 50
GRIDRES = '10.0'
NATIVE = '''
                                   {{
       
                                      "type": "rmsd",
       
                                      "Native": {{\n\
                                       "path":\n\
                                       "{}" }},\n\
       
                                      "selection": {{ "chains": {{ "names": [ "{}" ] }} }},\n\
       
                                      "includeHydrogens": false,\n\
       
                                      "doSuperposition": false,\n\
       
                                      "tag" : "ligandRMSD"\n\
       
                                   }},\n\
       
       
            '''
        
WATER_ENERGY =             '''
                            {{
                            "type": "bindingEnergy",\n\
                            "boundPartSelection": {{ "chains": {{ "names": ["{0}"] }} }},\n\
                            "tag": "water{0}"\n\
                            }},\n\
                           '''


UNBINDING = '''
            "modeMovingBox" : "unbinding",
            "exitCondition" : {{
                "type" : "metricMultipleTrajectories",
                "params" : {{
                       "metricCol" : {},
                       "exitValue" : {},
                       "condition" : "{}",
                       "numberTrajectories" : {}
            }}
            }},
            '''


DISTANCE_ATOMS =     '''
                     {{
                     "type":"com_distance",
                     "tag":"distance{2}",
                     "selection_group_1":{{
                     "atoms": {{ "ids":["{0}"]}}
                     }},
                     "selection_group_2":{{
                     "atoms": {{ "ids":["{1}"]}}
                     }}
                     }},
                     '''
BOX = '''

                "Box": {{
                    "type": "sphericalBox",
                    "radius": {0},
                    "fixedCenter": {1}
                }},
'''                     

WATER = '''
         "WaterPerturbation":
         {{
             "watersToPerturb": {{ "links": {{ "ids": [ {} ] }} }},
             "parameters":
             {{
                 {}
                 "temperature": {},
                 "numberOfStericTrials": {},
                 "overlapFactor": {},
                 "COMConstraintConstant": {}
             }},
             "waterSites": {}
         }}, 
'''

PCA = '''"preloadedModesIn" : "{}",'''


SELECTION_TO_PERTURB = '"selectionToPerturb" : { "chains" : { "names" : [ "$CHAIN" ] } },'
PERTURBATION = '''
          "Perturbation": {
                $BOX
                "perturbationType":"naive",
                "translationDirection": "steered",
                "rotationAngles": "nonCoupled",
                "parameters": {
                    "numberOfStericTrials": $STERIC_TRIALS,
                    "steeringUpdateFrequency": $STEERING,
                    "overlapFactor": $OVERLAP
                }   
                
            },
'''
BE = '''
                        { "type": "bindingEnergy",

                           "boundPartSelection": { "chains": { "names": ["$CHAIN"] } }

                        },
'''

SASA='''
                        { "type": "sasa",

                           "tag": "sasaLig",

                           "selection": { "chains": { "names": ["$CHAIN"] } }

                        },
'''



LIGAND = '"ligandResname" : "$LIG_RES",'

# PPI waters
water_O =  "HETATM {}  OW  HOH {}{:>4}     {}  1.00  0.00           O\n"
water_H1 = "HETATM {}  1HW HOH {}{:>4}     {}  1.00  0.00           H\n"
water_H2 = "HETATM {}  2HW HOH {}{:>4}     {}  1.00  0.00           H\n"
water = [water_O, water_H1, water_H2]


# METALS
metals = ['LI', 'BE', 'NA', 'MG', 'AL', 'K', 'CA', 'SC', 'TI', 'V', 'CR', 'MN', 'FE', 'CO', 'NI', 'CU', 'ZN', 'GA', 'RB', 'SR', 'Y', 'ZR', 'NB', 'MO', 'TC', 'RU', 'RH', 'PD', 'AG', 'CD', 'IN', 'SN', 'CS', 'BA', 'LA', 'CE', 'PR', 'ND', 'PM', 'SM', 'EU', 'GD', 'TB', 'DY', 'HO', 'ER', 'TM', 'YB', 'LU', 'HF', 'TA', 'W', 'RE', 'OS', 'IR', 'PT', 'AU', 'HG', 'TL', 'PB', 'BI', 'PO', 'FR', 'RA', 'AC', 'TH', 'PA', 'U', 'NP', 'PU', 'AM', 'CM', 'BK', 'CF', 'ES', 'FM', 'MD', 'NO', 'LR', 'RF', 'DB', 'SG', 'BH', 'HS', 'MT', 'DS', 'RG', 'CN', 'NH', 'FL', 'MC', 'LV']


#TEMPLATE KEYWORDS
#------------------

GLIDE_TEMPLATE = ["INPUT", "PRECISION"]

#RESTARTS:
#-----------

FIRST_RESTART = ["all",]
SECOND_RESTART = ["all", "adaptive"]
THIRD_RESTART = ["all", "adaptive", "pele"]
FOURTH_RESTART = ["all", "adaptive", "pele", "msm"] 

#PATHS
#-------

DIR = os.path.dirname(os.path.dirname(__file__))
ADAPTIVE_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), "Adaptive/clusterAdaptiveRun.py"))

#MESSAGES&ERRORS
#-----------------

CLUSTER_ERROR = "Number of cpus ({}) must be bigger than clusters ({})"
SYSTEM = "\n\t**Missing residues found {}\n\t**Gaps found {}\n\t**Metals found {}"



import os
import shutil
import AdaptivePELE.adaptiveSampling as adt
import PPP.main as ppp
from pele_platform.Utilities.Helpers import helpers
import pele_platform.Utilities.Parameters.pele_env as pele
import pele_platform.Utilities.Helpers.constraints as ct
import pele_platform.constants.constants as cs
import pele_platform.Utilities.Helpers.simulation as ad
import pele_platform.Utilities.Helpers.system_prep as sp
import pele_platform.Utilities.Helpers.missing_residues as mr
import pele_platform.Utilities.Helpers.randomize as rd
import pele_platform.Utilities.Helpers.helpers as hp
import pele_platform.Utilities.Helpers.Metals.metal_constraints as mc
import pele_platform.Utilities.Helpers.Metals.metal_polarisation as mp
import pele_platform.Adaptive.metrics as mt
import pele_platform.Utilities.Helpers.water as wt
import pele_platform.Analysis.plots as pt
import pele_platform.Utilities.Parameters.pele_env as pv
import pele_platform.Adaptive.ligand_parametrization as lg
import pele_platform.Adaptive.box as bx
import pele_platform.Adaptive.solvent as sv
import pele_platform.Adaptive.pca as pca
# import pele_platform.RNA.prep as pr


def run_adaptive(args: pv.EnviroBuilder) -> pv.EnviroBuilder:
    """
    Main function to prepare and launch simulation

    1) Crate working folders and inputs
    2) Prepare ligand and receptor
    3) Launch simulation
    4) Analyse simulation
    """

    # Build Folders and Logging and env variable that will contain
    # all main attributes of the simulation
    env = pele.EnviroBuilder()
    env.software = "Adaptive"
    env.build_adaptive_variables(args)
    env.create_files_and_folders()
    shutil.copy(args.yamlfile, env.pele_dir)

    if env.adaptive_restart and not env.only_analysis:
        with helpers.cd(env.pele_dir):
            adt.main(env.ad_ex_temp)
            env.logger.info("Simulation run successfully (:\n\n")

    elif not env.only_analysis:

        env.logger.info("System: {}; Platform Functionality: {}\n\n".format(env.residue, env.software))
        
        if env.perturbation:
            syst = sp.SystemBuilder.build_system(env.system, args.mae_lig, args.residue, env.pele_dir)
        else:
            syst = sp.SystemBuilder(env.system, None, None, env.pele_dir)
        
        env.logger.info("Prepare complex {}".format(syst.system))
           
        # Choose your own input
        
        # User specifies more than one input
        if env.input:
            env.inputs_simulation = []
            for input in env.input:
                input_path  = os.path.join(env.pele_dir, os.path.basename(input))
                shutil.copy(input, input_path)
                
                if env.no_ppp:
                    input_proc = input
                else:
                    input_proc = os.path.basename(ppp.main(input_path, env.pele_dir, output_pdb=["" , ],
                                charge_terminals=args.charge_ter, no_gaps_ter=args.gaps_ter,
                                constrain_smiles=env.constrain_smiles, ligand_pdb=env.ligand_ref)[0])
                env.inputs_simulation.append(input_proc)
                hp.silentremove([input_path])
            env.adap_ex_input = ", ".join(['"' + input + '"' for input in env.inputs_simulation]).strip('"')
        
        # Global exploration mode: Create inputs around protein
        elif args.full or args.randomize or args.ppi:
            ligand_positions, box_radius, box_center = rd.randomize_starting_position(env.ligand_ref, env.receptor,
                outputfolder=env.pele_dir, nposes=env.poses, test=env.test, user_center=env.center_of_interface, logger=env.logger)
            if not args.gpcr_orth:
                env.box_center = box_center
                env.box_radius = box_radius
            if env.no_ppp:
                receptor = syst.system
            else:
                receptor = ppp.main(syst.system, env.pele_dir, output_pdb=["" , ],
                            charge_terminals=args.charge_ter, no_gaps_ter=args.gaps_ter,
                            constrain_smiles=env.constrain_smiles, ligand_pdb=env.ligand_ref)[0]
            inputs = rd.join(receptor, ligand_positions, env.residue, output_folder=env.pele_dir)
            env.adap_ex_input = ", ".join(['"' + os.path.basename(input) + '"' for input in inputs]).strip('"')
            hp.silentremove(ligand_positions)
            # Parsing input for errors and saving them as inputs

        # Prepare System
        if env.no_ppp:
            missing_residues = []
            gaps = {}
            metals = {}
            env.constraints = ct.retrieve_constraints(env.system, gaps, metals, back_constr=env.ca_constr)
            if env.input:
                # If we have more than one input
                for input in env.input:
                    shutil.copy(input, env.pele_dir)
            else:
                shutil.copy(env.system, env.pele_dir)
        else:
            env.nonstandard.extend(hp.find_nonstd_residue(syst.system))
            env.system, missing_residues, gaps, metals, env.constraints = ppp.main(syst.system, env.pele_dir, output_pdb=["" , ], charge_terminals=args.charge_ter, no_gaps_ter=args.gaps_ter, mid_chain_nonstd_residue=env.nonstandard, skip=env.skip_prep, back_constr=env.ca_constr, constrain_smiles=env.constrain_smiles, ligand_pdb=env.ligand_ref)

        # Metal constraints
        if not args.no_metal_constraints:
            metal_constraints, env.external_constraints = mc.main(args.external_constraints, os.path.join(env.pele_dir, env.adap_ex_input.split(",")[0].strip().strip('"')), syst.system, permissive=env.permissive_metal_constr, all_metals=args.constrain_all_metals, external=env.external_constraints, logger=env.logger)
            env.external_constraints = hp.retrieve_constraints_for_pele(env.external_constraints, env.system)
            metal_constraints_json = hp.retrieve_constraints_for_pele(metal_constraints, env.system)
            env.external_constraints.extend(metal_constraints_json)
        else:
            env.external_constraints = hp.retrieve_constraints_for_pele(env.external_constraints, env.system)

        # Keep JSON ordered by having first title and then constraints
        if env.external_constraints:
            env.constraints = env.constraints[0:1] + env.external_constraints + env.constraints[1:]
        if env.remove_constraints:
            env.constraints = ""
        env.logger.info(cs.SYSTEM.format(missing_residues, gaps, metals))
        env.logger.info("Complex {} prepared\n\n".format(env.system))

        # Ligand parameters and simulation box
        if env.perturbation:
            ligand_params = lg.LigandParametrization(env)
            ligand_params.generate()
            box = bx.BoxSetter(env.box_center, env.box_radius, env.ligand_ref, env.logger)
            env.box = box.generate_json()
        else:
            env.box = ""

        # Parametrize missing residues
        for res, __, _ in missing_residues:
            if res != args.residue and res not in env.skip_ligand_prep:
                env.logger.info("Creating template for residue {}".format(res))
                with hp.cd(env.pele_dir):
                    mr.create_template(env, res)
                env.logger.info("Template {}z created\n\n".format(res))
    
        # Solvent parameters
        solvent = sv.ImplicitSolvent(env.solvent, env.obc_tmp, env.template_folder, env.obc_file, env.logger)
        solvent.generate()
        
        # Build PCA
        if env.pca_traj:
            pca_obj = pca.PCA(env.pca_traj, env.pele_dir)
            env.pca = pca_obj.generate(env.logger)
        
        ####### Add waters, if needed
        input_waters = [input.strip().strip('"') for input in env.adap_ex_input.split(",")]
        input_waters = [os.path.join(env.pele_dir, inp) for inp in input_waters]
        water_obj = wt.WaterIncluder(input_waters, env.n_waters, 
            user_waters=env.waters, ligand_perturbation_params=env.parameters, 
            water_center=args.water_center, water_radius=env.water_radius,
            allow_empty_selectors=env.allow_empty_selectors, water_temp=env.water_temp,
            water_trials=env.water_trials, water_overlap=env.water_overlap,
            water_constr=env.water_constr, test=env.test, water_freq=env.water_freq,
            ligand_residue=env.residue)
        water_obj.run()
        env.parameters = water_obj.ligand_perturbation_params

        #metrics
        metrics = mt.MetricBuilder()
        if args.atom_dist:
            env.metrics = metrics.distance_to_atom_json(env.system, args.atom_dist)
        else:
            env.metrics = ""
        if args.native:
            env.native = metrics.rsmd_to_json(args.native, env.chain)
        else:
            env.native = ""

        # metal polarisation
        if env.polarize_metals:
            mp.change_metal_charges(env.template_folder, env.forcefield, env.polarization_factor, env.system)

        ############Fill in Simulation Templates############
        # Fill in simulation templates
        adaptive = ad.SimulationBuilder(env.ad_ex_temp,
            env.pele_exit_temp, env.topology)
        adaptive.generate_inputs(env, water_obj)

    if env.analyse and not env.debug:
        report = pt.analyse_simulation(env.report_name, env.traj_name[:-4]+"_", 
            os.path.join(env.pele_dir, env.output), env.residue, cpus=env.cpus,
            output_folder=env.pele_dir, clustering=env.perturbation, mae=env.mae,
            nclusts=env.analysis_nclust, overwrite=env.overwrite, topology=env.topology,
            be_column=env.be_column, limit_column=env.limit_column, te_column=env.te_column, logger=env.logger)
        env.logger.info("Pdf summary report successfully written to: {}".format(report))
    return env



from dataclasses import dataclass
import glob
import os
import pele_platform.constants.constants as cs
import pele_platform.Utilities.Helpers.helpers as helpers
import pele_platform.Utilities.Helpers.calculatePCA4PELE as pc


@dataclass
class PCA:
    """
    Base class to perform pca analysis
    on user's receptor traj and generate
    the json input's code. Two input options:

    1) pca_traj = trajectory*.pdb
    2) pca_traj = [trajectory1.pdb, trajectory2.pdb, ...]
    """

    pca_traj: list
    pele_dir: str

    def generate(self, env) -> str:
        # Calculate pca and retrieve json format
        pdbs = self._retrieve_trajectories()
        pca_json = self._pca_to_json(pdbs, logger=env)
        return pca_json

    def _retrieve_trajectories(self) -> list:
        # Get pdb inputs to build pca
        if isinstance(self.pca_traj, str):
            pdbs = glob.glob(self.pca_traj)
        elif isinstance(self.pca_traj, list):
            pdbs = self.pca_traj
        return pdbs

    def _pca_to_json(self, pdbs, logger=None) -> str:
        # calculate pca over pdb and get json
        pdbs_full_path = [os.path.abspath(pdb) for pdb in pdbs]
        with helpers.cd(self.pele_dir):
            pca = pc.main(pdbs_full_path, logger=logger)
        self.pca = cs.PCA.format(pca)
        return self.pca



import os
import sys
import subprocess
import argparse
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))
import pele_platform.constants.constants as cs

class SystemBuilder(object):
    def __init__(self, receptor, ligand, residue, pele_dir):
        self.receptor = receptor
        self.ligand = ligand
        self.residue = residue
        self.pele_dir = pele_dir
        self.system = self.receptor

    @classmethod
    def build_system(cls, receptor, ligand, residue, pele_dir, output=False):
        SPYTHON = os.path.join(cs.SCHRODINGER, "utilities/python")
        if not os.path.exists(SPYTHON):
            SPYTHON = os.path.join(cs.SCHRODINGER, "run")
        system = cls(receptor, ligand, residue, pele_dir)
        system.receptor, system.lig_ref = system.retrieve_receptor(output=output)
        system.lig = ligand if ligand else "{}.mae".format(residue)
        my_env = os.environ.copy()
        my_env["PYTHONPATH"] = ''
        my_env["SCHRODINGER_PYTHONPATH"]=os.path.join(cs.SCHRODINGER, "internal/lib/python2.7/site-packages/")
        my_env["SCHRODINGER"]=cs.SCHRODINGER
        subprocess.call("{} {} {} {}".format(SPYTHON, __file__, system.lig_ref, pele_dir).split(), env=my_env)
        system.residue = residue
        return system

    def build_complex(self):
        """
            From the receptor and ligand in pdb build
            another pdb with the whole complex
        """
        complex_content = []

        name = os.path.basename(os.path.splitext(self.receptor)[0])
        self.complex = os.path.join(self.pele_dir, "{}_complex.pdb".format(name))

        try:
            with open(self.receptor, 'r') as pdb_file:
                receptor_text = [line for line in pdb_file if line.startswith("ATOM") or line.startswith("HETATM") or line.startswith("TER")]
            with open(self.lig_ref, 'r') as pdb_file:
                ligand_text = [line for line in pdb_file if line.startswith("HETATM")]
        except IOError:
            raise IOError("Receptor or ligand not found. Check your ligand residue/chain name or input files")

        if not receptor_text  or not ligand_text:
            raise ValueError("The ligand_pdb was not properly created check your mae file")

        complex_content.extend(receptor_text + ["TER\n"] + ligand_text + ["END"])

        with open(self.complex, 'w') as fout:
            fout.write("".join(complex_content))

        return self.complex

    def convert_mae(self):
        """
           Desciption: From each structure retrieve
           a .mae file of the ligand in the receptor.
           Output:
                structure_mae: ligand
                res = residue
        """

        for structure in st.StructureReader(self.lig_ref):
            for residue in structure.residue:
                res = residue.pdbres.strip()
            str_name = "{}".format(res)
            try:
                structure.write(str_name + ".mae")
            except ValueError:
                str_name = "{}".format(res)
            finally:
                structure.write(str_name + ".mae")
                structure_mae = "{}.mae".format(str_name)
        return structure_mae, res

    def retrieve_receptor(self, output=False):
        """
        This function returns receptor of the complex of interest.
        :param complex: system format pdb
        :output: receptor text
        """
        ligand = output if output else os.path.join(self.pele_dir, "ligand.pdb")
        receptor =  os.path.join(self.pele_dir, "receptor.pdb")

        with open(self.receptor, 'r') as pdb_file:
            receptor_text = [line for line in pdb_file if line.startswith("ATOM") or "TER" in line or (line.startswith("HETATM") and line[17:20].strip() != self.residue)]
        with open(self.receptor, 'r') as pdb_file:
            residue = None
            ligand_text = []
            for line in pdb_file:
                if line[17:20].strip() == self.residue:
                    if residue is None:
                        residue = line[22:26]
                        chain = line[20:22]
                        ligand_text.append(line)
                    else:
                        if residue == line[22:26] and chain == line[20:22]:
                            ligand_text.append(line)

        if not receptor_text  or not ligand_text:
            raise ValueError("Something went wrong when extracting the ligand. Check residue&Chain on input")
        with open(ligand, "w") as fout:
            fout.write("".join(ligand_text+["END"]))
        with open(receptor, "w") as fout:
            fout.write("".join(receptor_text+["END"]))

        return "".join(receptor_text), ligand

def convert_pdb(mae_file, output_dir):
    from schrodinger import structure as st
    for structure in st.StructureReader(mae_file):
        structure.write(output_dir)


def convert_mae(pdb, output_dir="."):
    """
        Desciption: From each structure retrieve
        a .mae file of the ligand in the receptor.
        Output:
             structure_mae: ligand
             res = residue
    """
    from schrodinger import structure as st
    for structure in st.StructureReader(pdb):
        for residue in structure.residue:
            res = residue.pdbres.strip()
            str_name = os.path.abspath(os.path.join(output_dir, "{}".format(res)))
            try:
                structure.write(str_name + ".mae")
            except ValueError:
                str_name = "{}".format(res)
            finally:
                structure.write(str_name + ".mae")
                structure_mae = "{}.mae".format(str_name)

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("ligand", help="ligand input file to convert")
    parser.add_argument("output_dir", help="output directory to dump the converted file")
    parser.add_argument("--mae", action="store_true", help="Whether to convert to mae (--mae) or pdb (not --mae)")
    args = parser.parse_args()
    return args.ligand, args.output_dir, args.mae




if __name__ == "__main__":
    input_file, output_dir, ligand_mae = parse_args()
    if ligand_mae:
        convert_pdb(input_file, output_dir)
    else:
        convert_mae(input_file, output_dir)
        



import pele_platform.Adaptive.metrics as mt


class FragMetrics(mt.MetricBuilder):


	def __init__(self, args):
            if args.atom_dist:
                self.metrics = self.distance_to_atom_json(self.system, args.atom_dist)
            else:
                self.metrics = ""
            if args.native:
                self.native = self.rsmd_to_json(args.native, self.chain)
            else:
                self.native = ""



import pele_platform.Utilities.Helpers.constraints as cst
import pele_platform.constants.constants as cs


class FragOpcionalParameters():

    def __init__(self, args):
        # SIMULATION CONTROL
        self.frag_run = args.frag_run
        # CONSTRAINTS
        self.constraints = cst.retrieve_constraints(self.core, {}, {}, self.ca_constr)
        # CHAIN
        self.chain_core = args.chain_core if args.chain_core else self.simulation_params.get("chain_core", "L")
        # BOX
        self.box = cs.BOX.format(self.box_radius, self.box_center) if  self.box_radius else ""
        # LIGAND
        self.gridres = args.gridres
        self.plop_path = "PlopRotTemp_S_2017/ligand_prep.py"
        # OUTPUT
        self.criteria = args.frag_criteria if args.frag_criteria else self.simulation_params.get(
            "frag_criteria", "Binding Energy")
        self.output_folder = args.frag_output_folder if args.frag_output_folder else self.simulation_params.get(
            "frag_output_folder", "growing_steps")
        self.cluster_folder = args.frag_cluster_folder if args.frag_cluster_folder else self.simulation_params.get(
            "frag_cluster_folder", "clustering_PDBs")
        # OTHERS
        self.distcont = 4
        self.threshold = 0.3
        self.condition = "min"
        self.metricweights = "linear"
        self.nclusters = 5
        self.min_overlap = 0.5
        self.max_overlap = 0.7
        self.frag_chain = "L"
        self.banned = None
        self.limit = None
        self.rename = False
        self.threshold_clash = 1.7
        self.translation_high = 0.05
        self.translation_low = 0.02
        self.rotation_high = 0.1
        self.rotation_low = 0.05
        self.explorative = False
        self.frag_radius = 10
        self.sampling_control = None
        self.only_prepare = False
        self.only_grow = False
        self.no_check = True
        self.cleanup = args.cleanup if args.cleanup else False







class LigandSizeExceed(Exception):
    pass

class MissResidueFlag(Exception):
    pass

class SameMolecule(Exception):
    pass

class ExecutableNotInPath(Exception):
    pass

class OldPythonVersion(Exception):
    pass

class LigandPreparationError(Exception):
    pass

class NoGeometryAroundMetal(Exception):
    pass

class EnvVariableNotFound(Exception):
    pass

class WrongAtomSpecified(Exception):
    pass

class NotCenterOfWaterBox(Exception):
    pass

class OutInError(Exception):
    pass

class TemplateFileNotFound(Exception):
    pass

class RotamersFileNotFound(Exception):
    pass

class WrongAtomStringFormat(Exception):
    pass

class WrongYamlFile(Exception):
    pass



import os
import pele_platform.constants.constants as cs


class SimulationPaths():


    def __init__(self, args):
        self.working_folder_paths(args)
        self.ligand_paths(args)
        self.complex_paths(args)
        self.solvent_paths(args)
        self.box_paths(args)


    def working_folder_paths(self, args):
        if self.no_ppp:
            self.system_fix = os.path.join(self.pele_dir, os.path.basename(self.system))
        else:
            self.system_fix = os.path.join(self.pele_dir, "{}_processed.pdb".format(os.path.splitext(os.path.basename(self.system))[0]))
        self.adap_ex_input = os.path.basename(self.system_fix)

        self.pele_temp = os.path.join(self.pele_dir, "pele.conf")
        self.adap_l_input = "{}/initial_*"
        self.adap_l_output = os.path.join(self.pele_dir, "output_pele")
        self.ad_l_temp = os.path.join(self.pele_dir, "adaptive_long.conf")

    def ligand_paths(self, args):
        self.ligand_ref = os.path.join(self.pele_dir, "ligand.pdb")
        self.rotamers_folder = os.path.join(self.pele_dir, "DataLocal/LigandRotamerLibs/")
        self.template_folder = os.path.join(self.pele_dir, "DataLocal/Templates/{}/HeteroAtoms/".format(self.forcefield))

    def complex_paths(self, args):
        self.receptor = os.path.join(self.pele_dir, "receptor.pdb")
        self.topology = None if self.pdb else os.path.join(self.pele_dir, self.output, "topologies/topology_0.pdb")

    def solvent_paths(self, args):
        self.obc_tmp = os.path.join(cs.DIR, "Templates/solventParamsHCTOBC.txt")
        self.obc_file = os.path.join(self.pele_dir, "DataLocal/OBC/solventParamsHCTOBC.txt")

    def box_paths(self, args):
        self.box_temp = os.path.join(self.pele_dir, "box.pdb")



from dataclasses import dataclass
import pele_platform.Adaptive.simulation as si
import pele_platform.Utilities.Helpers.helpers as hp
import pele_platform.Utilities.Parameters.pele_env as pv
import pele_platform.Errors.custom_errors as ce


@dataclass
class OutInLauncher():

    args: pv.EnviroBuilder

    def run_gpcr_simulation(self) -> pv.EnviroBuilder:
        #Set parameters for gpcr and launch simulation
        self._check_mandatory_fields()
        self._set_parameters()
        simulation_parameters = si.run_adaptive(self.args)
        return simulation_parameters


    def _check_mandatory_fields(self):
        COMPULSORY_FLAGS = ["final_site", "initial_site"]
        for flag in COMPULSORY_FLAGS:
            if getattr(self.args, flag) is None:
                raise ce.OutInError(f"flag {flag} must be specified for out_in package")

        

    def _set_parameters(self) -> None:
        # Set box and initial ligand position
        self.final_site = self.args.final_site
        self.initial_site = self.args.initial_site
        self.args.center_of_interface = self.initial_site
        box_center, box_radius = hp.retrieve_box(
    self.args.system, self.initial_site, self.final_site,
    weights=[0.35, 0.65])
        self.args.box_center = self.args.box_center if self.args.box_center else box_center
        self.args.box_radius = self.args.box_radius if self.args.box_radius else box_radius
        self.args.randomize = True



from dataclasses import dataclass
import pele_platform.Checker.main as ck
import pele_platform.Frag.simulation as fr
import pele_platform.Adaptive.simulation as ad
from pele_platform.Allosteric.main import run_allosteric
import pele_platform.gpcr.main as gpcr
import pele_platform.out_in.main as outin
from pele_platform.PPI.main import run_ppi
import pele_platform.Utilities.Parameters.pele_env as pv
import argparse


@dataclass
class Launcher:

    _args: argparse.ArgumentParser
    frag: str="frag"
    ppi: str="PPI"
    allosteric: str="allosteric"
    gpcr_orth: str="gpcr_orth"
    out_in: str="out_in"
    adaptive: str="adaptive"

    def launch(self) -> pv.EnviroBuilder:
        # Launch package from input.yaml
        self._define_package_to_run()
        job_variables = self.launch_package(self._args.package, no_check=self._args.no_check)
        return job_variables

    def launch_package(self, package: str, no_check=False) -> pv.EnviroBuilder:
        # Launch package from API
        if not no_check:
            ck.check_executable_and_env_variables(self._args)
        if package == self.adaptive:
            job_variables = ad.run_adaptive(self._args)
        elif package == self.gpcr_orth:
            job_variables = gpcr.GpcrLauncher(self._args).run_gpcr_simulation()
        elif package == self.out_in:
            job_variables = outin.OutInLauncher(self._args).run_gpcr_simulation()
        elif package == self.allosteric:
            job_variables = run_allosteric(self._args)
        elif package == self.ppi:
            job_variables = run_ppi(self._args)
        elif package == self.frag:
            # Set variables and input ready
            job_variables = fr.FragRunner(self._args).run_simulation()
        return job_variables

    def _define_package_to_run(self) -> None:
        # Define package being run from input.yaml flags
        if self._args.frag_core:
            self._args.package = self.frag
        elif self._args.ppi:
            self._args.package = self.ppi
        elif self._args.allosteric:
            self._args.package = self.allosteric
        elif self._args.gpcr_orth:
            self._args.package = self.gpcr_orth
        elif self._args.out_in:
            self._args.package = self.out_in
        else: 
            self._args.package = self.adaptive



import os


def prepare_structure(protein_file, ligand_pdb, chain, remove_water=False):
    
    to_remove = []

    # remove additional protein chains and all water molecules
    with open(protein_file, "r") as file:
        lines = file.readlines()

        for line in lines:
            if ((line.startswith("ATOM") or line.startswith("HETATM")) and line[21:22].strip() not in chain) or \
                    line.startswith("END") or line.startswith("TER") or line.startswith("CONECT"):
                to_remove.append(line)
            if remove_water:
                if (line.startswith("ATOM") or line.startswith("HETATM")) and line[17:20].strip() == "HOH":
                    to_remove.append(line)

        protein = [line for line in lines if line not in to_remove]
    
    # read in ligand file
    ligand = []
    with open(ligand_pdb, "r") as ligand_file:
        lines = ligand_file.readlines()
        for line in lines:
            if line.startswith("ATOM") or line.startswith("HETATM"):
                ligand.append(line)
    
    new_protein_file = os.path.basename(protein_file).replace(".pdb", "_prep.pdb")
    new_protein_file = os.path.abspath(new_protein_file)
    
    # join protein and ligand PDBs into new file
    with open(new_protein_file, "w+") as file:
        for line in protein:
            file.write(line)
        file.write("\n")
        for line in ligand:
            file.write(line)
 
    return new_protein_file



import sys
import pele_platform.Errors.custom_errors as ce

def check_python_version():
    #Check python version is > 3.7
    python_version = sys.version_info
    if sys.version_info < (3, 7):
        raise ce.OldPythonVersion("Pele Platform requieres Python>3.6. Used is {}.{}".format(
        python_version.major, python_version.minor))



import argparse
import os
import re
import time
from pele_platform.Utilities.Helpers.yaml_parser import YamlParser
import pele_platform.Checker.valid_flags as vf


def prep_complex(complex, input_file="input.yaml", prep_output="", final_output="", debug=False):

    input_path = os.path.abspath(complex)
    if not prep_output:
        prep_output = os.path.basename(complex.replace(".pdb", "_prep.pdb"))
    if not final_output:
        final_output = os.path.basename(complex.replace(".pdb", "_final.pdb"))
    schrodinger_path = "$SCHRODINGER/utilities/prepwizard"
    
    # Run Protein Preparation Wizard - delete waters, fill missing loops and side chains
    wizard_command = "{} -fillloops -fillsidechains -delwater_hbond_cutoff 5 {} {}".format(schrodinger_path, input_path, prep_output)
    if not debug:
        os.system(wizard_command)

    # Get input.yaml
    yaml = os.path.abspath(input_file)
    args = YamlParser(yaml, vf.VALID_FLAGS_PLATFORM)._parse_yaml()
    input_resname = args["resname"]

    # Read in PDB file
    while not os.path.exists(prep_output):
        time.sleep(10)

    with open(prep_output, "r") as pdb:
        lines = pdb.readlines()
        all_chains = [line[21:22].strip() for line in lines if line.startswith("ATOM") or line.startswith("HETATM")]
        lig_chains = [line[21:22].strip() for line in lines if line[17:20] == input_resname]
        lig_atomnames = [line[12:16].strip() for line in lines if line [17:20] == input_resname]
    
    # Change ligand chain and atom names, if not unique
    lig_occ = all_chains.count(lig_chains[0])
    lig_length = len(lig_chains)
    
    unique_chain = True if lig_occ == lig_length else False
    unique_atomname = True if len(set(lig_atomnames)) == len(lig_atomnames) else False


    with open(prep_output, "r") as file:
        lines = file.readlines()
        dict = {} 
        with open(final_output, "w") as final_output:
            for line in lines:
                if unique_chain == False:
                    if line[17:20] == input_resname:
                        repl_chain = "Z"
                        line = line.replace(line[21],repl_chain)
                if unique_atomname == False:
                    if line[17:20] == input_resname:
                        pattern = r"[^0-9\s]"
                        letter = re.search(pattern, line[12:16])[0].strip()
                        if letter not in dict.keys():
                            dict[letter] = 1
                        else:
                            dict[letter] += 1
                        replacement = "{}{}".format(letter, dict[letter])
                        if len(replacement) != 4:
                            replacement = replacement + " "*(4 - len(replacement))
                        line = line.replace(line[12:16], replacement)
                final_output.write(line)

    return final_output.name


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--complex", type=str, required=True, help="Protein-ligand complex (PDB file)")
    parser.add_argument("--input", type=str, required=True, help="YAML file to launch simulation (input.yaml)") 
    args = parser.parse_args()
    return os.path.abspath(args.complex), os.path.abspath(args.input)


if __name__ == "__main__":
    complex, input_file = parse_args()
    prep_complex(complex, input_file)



class GenerativeModel:

    def __init__(self, core, resname, iterations, only_grow=True, only_rank=False):
        self.core = core
        self.iterations = iterations
        self.only_grow = only_grow
        self.only_rank = only_rank
        self.resname = resname

    def run(self):
        from growai import grow
        return grow.main(self.core, self.resname, self.only_grow, self.only_rank, self.iterations)


import os
import pele_platform.Adaptive.simulation as si

DIR = os.path.dirname(os.path.abspath(__file__))


def launch_simulation(original_yaml):
    job_parameters = si.run_adaptive(original_yaml)
    return job_parameters



import pele_platform.constants.pele_params as pcs

"""
Description of the file: This file specifies what files to use
for each job.

"""
    
def retrieve_software_settings(args):

        SOFTWARE_CONSTANTS = {
                 "simulation_params" : {
                 "standard": {"params": pcs.FRAG, "COMligandConstraint": 1, 
                 "steric_trials": 100, "overlap_factor": 0.5,
                 "sidechain_freq": 1, "anm_freq": 2}
        }
        }
        
        software_setings = SOFTWARE_CONSTANTS
        type_simulation = "standard"
        software_setings["simulation_params"] = software_setings["simulation_params"].get(type_simulation, {})
        return software_setings




class AllostericParams(object):


    def generate_allosteric_params(self, args):
        self.allosteric = args.allosteric
        self.n_components = args.n_components if args.n_components else  self.simulation_params.get("n_components", 10)



from fpdf import FPDF
import os
import numpy as np


def create_report(plots, clusters, top_poses, best_energies, output="simulation_summary.pdf"):


    pdf = FPDF()
    pdf.add_page() 

    # Title
    pdf.set_font('Arial', 'B', 15)
    idx_25 = int(len(best_energies)*0.25)+1
    mean_25_quartile = np.mean(sorted(best_energies)[:idx_25])
    pdf.cell(0, 10, "Score {} kcal/mol".format(mean_25_quartile), align='C')

    pdf.ln(40) #4 linebreaks


    #Plots
    pdf.set_font('Arial', 'B', size=12)
    pdf.cell(0, 10, "Plots", align='C')
    pdf.set_font('Arial', size=10)
    for i, plot in enumerate(plots):
        if i%2 == 0 and i!=0:
            pdf.ln(1000000) #pagebreak
        pdf.ln(10)
        name = os.path.basename(plot)[:-4]
        pdf.cell(0, 50, name, align='C')
        pdf.ln(30)
        pdf.image(plot, x=70, w=83)

    pdf.ln(10000000) #page break


    #Top poses
    pdf.set_font('Arial', 'B', size=12)
    pdf.cell(0, 10, "Top poses", align='C')
    pdf.set_font('Arial', size=10)
    top_poses_ordered = np.array(top_poses)[np.argsort(best_energies)[:len(top_poses)]]
    for i, poses in enumerate(top_poses_ordered[0:20]):
        if i == 0:
            pdf.ln(10)
        pdf.cell(0, 50, os.path.basename(poses), align='C')
        pdf.ln(10)

    pdf.ln(10000000) #page break

    #Clusters
    pdf.set_font('Arial', 'B', size=12)
    pdf.cell(0, 10, "Clusters", align='C')
    pdf.set_font('Arial', size=10)
    for i, plot in enumerate(clusters):
        if i%2 == 0 and i!=0:
            pdf.ln(60)
        pdf.ln(10)
        name = os.path.basename(plot)[:-4]
        pdf.cell(0, 50, name, align='C')
        pdf.ln(30)
        pdf.image(plot, x=70, w=83)

    #Output report    
    pdf.output(output, 'F')
    return output



import pele_platform.constants.constants as cs

class PCAParams(object):


    def generate_pca_params(self, args):
        self.pca = cs.PCA.format(args.pca) if args.pca else ""
        self.pca_traj = args.pca_traj
        if self.pca or self.pca_traj:
            self.anm_direction = args.anm_direction if args.anm_direction else self.simulation_params.get("anm_direction", "oscillate")
            self.anm_mix_modes = args.anm_mix_modes if args.anm_mix_modes else self.simulation_params.get("anm_mix_modes", "doNotMixModes")
            self.anm_picking_mode = args.anm_picking_mode if args.anm_picking_mode else self.simulation_params.get("anm_picking_mode", "LOWEST_MODE")
            self.anm_displacement = args.anm_displacement if args.anm_displacement else self.simulation_params.get("anm_displacement", 2.0)
            self.anm_modes_change = args.anm_modes_change if args.anm_modes_change else self.simulation_params.get("anm_modes_change", 50)
            self.anm_num_of_modes = args.anm_num_of_modes if args.anm_num_of_modes else self.simulation_params.get("anm_num_of_modes", 1)
            self.anm_relaxation_constr = args.anm_relaxation_constr if args.anm_relaxation_constr else self.simulation_params.get("anm_relaxation_constr", 2)
            self.remove_constraints = args.remove_constraints if args.remove_constraints is not None else self.simulation_params.get("remove_constraints", True)
            self.anm_freq = args.anm_freq if args.anm_freq else self.simulation_params.get("anm_freq", 1)



from pele_platform.Utilities.Helpers import water as wt


class FragWaterParams():

    def __init__(self, pdb):
        self.pdb = pdb
        self.initialize_waters()

    def initialize_waters(self):
        self.water_object = wt.WaterIncluder([self.pdb], self.n_waters,
            user_waters=self.waters, ligand_perturbation_params=self.parameters,
            water_center=self.water_center, water_radius=self.water_radius,
            allow_empty_selectors=self.allow_empty_selectors, water_temp=self.water_temp,
            water_trials=self.water_trials, water_overlap=self.water_overlap,
            water_constr=self.water_constr, test=self.test, water_freq=self.water_freq,
            ligand_residue=self.residue)
        self.water_object.run()
    



from pele_platform.Frag.parameters import  files, simulation, opcionals, water, metrics
import pele_platform.Utilities.Parameters.pele_env as pele



class FragParameters(pele.EnviroBuilder, water.FragWaterParams,
    files.FragInputFiles, simulation.FragSimulationParameters,
    metrics.FragMetrics):

    def __init__(self, args):
        self.software = "Frag"

        # Platform common variables
        self.build_frag_variables(args)

        #Water Parameters
        water.FragWaterParams.__init__(self, args.frag_core)

        #Frag Input_files parameters
        files.FragInputFiles.__init__(self, args)
        
        # Frag Simulation Parameters
        simulation.FragSimulationParameters.__init__(self, args)

        #Frag Metric Parameters
        metrics.FragMetrics.__init__(self, args)

        # Frag Optional Parameters
        opcionals.FragOpcionalParameters.__init__(self, args)

        # Keep initial arguments
        self.args = args






import os
import PlopRotTemp as plop
import pele_platform.constants.constants as cs
import pele_platform.Errors.custom_errors as ce

try:
    import subprocess32 as subprocess
except ImportError:
    import subprocess
except SyntaxError:
    import subprocess


def parametrize_miss_residues(env, resname=None, ligand=None):
    resname = env.residue if not resname else resname
    SPYTHON = os.path.join(cs.SCHRODINGER, "utilities/python")
    if not os.path.exists(SPYTHON):
        SPYTHON = os.path.join(cs.SCHRODINGER, "run")
    file_path = os.path.join(os.path.dirname(plop.__file__), "main.py")
    options = retrieve_options(env)
    templatedir = os.path.join(env.pele_dir, "DataLocal/Templates/OPLS2005/HeteroAtoms")
    rotamerdir = os.path.join(env.pele_dir, "DataLocal/LigandRotamerLibs")  
    my_env = os.environ.copy()
    my_env["SCHRODINGER_PYTHONPATH"]=os.path.join(cs.SCHRODINGER, "internal/lib/python2.7/site-packages/")
    my_env["SCHRODINGER"]=cs.SCHRODINGER
    env.logger.info("Running Plop")
    ligand = ligand if ligand else env.lig
    env.logger.info("{} {} {} {} --outputname {} --templatedir {} --rotamerdir {}".format(SPYTHON, file_path, options, ligand, resname, templatedir, rotamerdir))
    try:
        subprocess.check_output("{} {} {} {} --outputname {} --templatedir {} --rotamerdir {}".format(SPYTHON, file_path, options, ligand, resname, templatedir, rotamerdir).split(), env=my_env)
    except subprocess.CalledProcessError: 
        raise ce.LigandPreparationError(
"\n\nLigand preparation failed.\n\
##############################\n\n\
1) Check there are no spaces in the ligand atom name and that \
the inputted ligand has a valid structure.\n\
2)Also, if LICENSE -1 FAIL is found on the output please point out to schrodinger licenses by either doing:\n\
\t - export SCHROD_LICENSE_FILE=/path/to/folder/with/static/license\n\
\t - export LM_LICENSE_FILE=/path/to/folder/with/server/license/")


def retrieve_options(env):
    """
    Retrieve PlopRotTemp options from input arguments
    """

    options = []
    if env.core != -1:
        options.extend(["--core {}".format(env.core)])
    if env.mtor != 4:
        options.extend(["--mtor {}".format(env.mtor)])
    if env.n != 1000:
        options.extend(["--n {}".format(env.n)])
    if env.forcefield != "OPLS2005":
        options.extend(["--force {}".format(env.forcefield)])
    if env.mae_lig:
        options.extend(["--mae_charges"])
    if env.gridres != 10:
        options.extend(["--gridres {}".format(env.gridres)])
    return " ".join(options)




import pele_platform.constants.constants as cs
import pele_platform.Errors.custom_errors as ce
from Bio.PDB import PDBParser, NeighborSearch, Selection, Vector, vectors
import itertools
import numpy as np


def find_metals(protein_file):

    # read in the protein file
    parser = PDBParser()
    structure = parser.get_structure("protein", protein_file)

    # find metals
    metals = []
    for chain in structure.get_chains():
        for residue in chain.get_residues():
            for atom in residue.get_atoms():
                if atom.element in cs.metals:
                    metals.append([atom, residue, chain])
    return metals, structure


def map_constraints(protein_file, original_input, original_constraints):
    
    atoms = []
    new_atoms = []

    # get lines from actual input
    with open(protein_file, "r") as input_file:
        input_lines = input_file.readlines()

    # get constraints coords from original input file
    with open(original_input, "r") as file:
        lines = file.readlines()
        
        for orig in original_constraints:
            try:
                k, dist, atom1, atom2 = orig.split("-")
            except ValueError: #If more than one atom
                continue
            atoms.extend([atom1, atom2])

        for atom in atoms:
            chain, resnum, atom_name = atom.split(":")
            
            for line in lines:
                if (line.startswith("HETATM") or line.startswith("ATOM")) and line[21].strip() == chain.strip() and line[22:26].strip() == resnum.strip() and line[12:16].strip() == atom_name.strip():
                    coords = line[30:54].split()
                    for l in input_lines:
                        if l[30:54].split() == coords:
                            new_atom_name = l[12:16].strip()
                            new_resnum = l[22:26].strip()
                            new_chain = l[21].strip()
                            new_atoms.append([chain, new_chain, resnum, new_resnum, atom_name, new_atom_name])
    
    before = ["{}:{}:{}".format(i[0],i[2],i[4]) for i in new_atoms]
    after = ["{}:{}:{}".format(i[1], i[3], i[5]) for i in new_atoms]
    
    for j in range(len(original_constraints)):
        for b, a in zip(before, after):
            original_constraints[j] = original_constraints[j].replace(b, a)

    return original_constraints


def angle_classification(combinations, permissive):

    # angle classification
    ang_180 = []
    ang_90 = []
    ang_109 = []
    coordinated_atoms = []

    if permissive:
        lower = 0.65
        upper = 1.35
    else:
        lower = 0.8
        upper = 1.2

    for c in combinations:
        a = c[2]
        if 180 * lower <= a <= 180 * upper:
            ang_180.append(c)
        if 90 * lower <= a <= 90 * upper:
            ang_90.append(c)
        if 109.5 * lower <= a <= 109.5 * upper:
            ang_109.append(c)
    
    # check geometries
    if len(ang_180) == 3 and len(ang_90) == 12:
        geo = "octahedral"
        coordinated_atoms.extend(ang_180)
        coordinated_atoms.extend(ang_90)
    elif len(ang_180) == 2 and len(ang_90) == 4:
        geo = "square planar"
        coordinated_atoms.extend(ang_180)
        coordinated_atoms.extend(ang_90)
    elif len(ang_109) == 6:
        geo = "tetrahedral"
        coordinated_atoms.extend(ang_109)
    else:
        geo = None
    
    return geo, coordinated_atoms


def find_geometry(metals, structure, permissive=False, all_metals=False, external=None, logger=None):

    # check metal contacts
    output = []
    checked_metals = []
    structure_list = Selection.unfold_entities(structure, "A")

    for metal in metals:

        # search distance based on metal type
        if metal[0].element == "YB":
            dist = 3.5
        elif metal[0].element == "K":
            dist = 3.3
        else:
            dist = 2.9 
        
        metal_str = "{}:{}:{}".format(metal[2].id, metal[1].get_id()[1], metal[0].name)
        in_ext = []

        for i in external:
            if metal_str in i:
                in_ext = True
        
        if not in_ext and list(metal[0].coord) not in checked_metals:
            coords = metal[0].coord

            contacts = []

            for chain in structure.get_chains():

                for residue in chain.get_residues():
                    contacts_atoms = NeighborSearch(structure_list).search(coords, dist, "A")
                    # exclude self-contacts, carbons and hydrogens
                    excluded_contacts = cs.metals + ['C', 'H']
                    contacts_atoms = [c for c in contacts_atoms if c.element not in excluded_contacts]

                    for atom in contacts_atoms:
                        if residue in chain.get_residues() and atom in residue.get_atoms():
                            contacts.append([atom, residue, chain])

            combinations = list(itertools.combinations(contacts, 2))
            combinations = [list(c) for c in combinations]
            
            # get all atom - metal - atom angles
            for c in combinations:
                vi = Vector(c[0][0].coord)
                vj = Vector(c[1][0].coord)
                angle = vectors.calc_angle(vi, coords, vj) * 180 / np.pi
                c.append(angle)

            geo, coordinated_atoms = angle_classification(combinations, False)

            if geo is None and permissive:
                geo, coordinated_atoms = angle_classification(combinations, True)
                
                if geo is None and all_metals and combinations:

                    geo, coordinated_atoms = angle_classification(combinations, True)
                    if geo:
                        logger.info("Found {} geometry around {} (residue {}). Adding constraints.".format(geo, metal[0].name, metal[1].get_id()[1]))
                        checked_metals.append(list(metal[0].coord))
                    else:
                        coordinated_atoms = combinations
                        checked_metals.append(list(metal[0].coord))
                        geo = "no"
                        logger.info("Found {} geometry around {} (residue {}). Adding constraints to all atoms within {}A of the metal.".format(geo, metal[0].name, metal[1].get_id()[1], dist))
                
                elif geo is None and not all_metals:
                    raise ce.NoGeometryAroundMetal("Failed to determine geometry around {} (residue {}). Add constraints manually or set 'constrain_all_metals: true' to constrain all atoms within {}A of the metal.".format(metal[0].name, metal[1].get_id()[1], dist))

                elif geo is None and all_metals and not combinations:
                    logger.info("No atoms coordinated to {} (residue {}).".format(metal[0].name, metal[1].get_id()[1]))

                elif geo:
                    checked_metals.append(list(metal[0].coord))
                    logger.info("Found {} geometry around {} (residue {}). Adding constraints.".format(geo, metal[0].name, metal[1].get_id()[1]))

            elif geo is None and all_metals and combinations:
                geo, coordinated_atoms = angle_classification(combinations, True)
                
                if geo is None:
                    geo = "no"
                    coordinated_atoms = combinations
                    checked_metals.append(list(metal[0].coord)) 
                    logger.info("Found {} geometry around {} (residue {}). Adding constraints to all atoms within {}A of the metal.".format(geo, metal[0].name, metal[1].get_id()[1], dist))

                else:
                    logger.info("Found {} geometry around {} (residue {}). Adding constraints.".format(geo, metal[0].name, metal[1].get_id()[1]))

            elif geo is None and all_metals and not combinations:
               logger.info("No atoms coordinated to {} (residue {}).".format(metal[0].name, metal[1].get_id()[1]))

            elif geo is None and not all_metals and not permissive:

                if not combinations:
                    print("No atoms coordinated to {} (residue {}).".format(metal[0].name, metal[1].get_id()[1]))
                else:
                    geo, coordinated_atoms = angle_classification(combinations, True)
                                                    
                    if geo is None:
                        geo = "no"
                        coordinated_atoms = combinations
                        checked_metals.append(list(metal[0].coord)) 
                        print("Found {} geometry around {} (residue {}). Adding constraints to all atoms within {}A of the metal.".format(geo, metal[0].name, metal[1].get_id()[1], dist))
            else:
                checked_metals.append(list(metal[0].coord))
                logger.info("Found {} geometry around {} (residue {}). Adding constraints.".format(geo, metal[0].name,
                                                                                             metal[1].get_id()[1]))

            # format string
            yaml_string = "{}-{}-{}:{}:{}-{}:{}:{}"
            spring_const = 50

            string_atoms = []
            for c in coordinated_atoms:
                atom1, atom2, angle = c

                if atom1 not in string_atoms:
                    string_atoms.append(atom1)
                if atom2 not in string_atoms:
                    string_atoms.append(atom2)

            for atom in string_atoms:
                atomname1 = atom[0].name
                resnum1 = atom[1].get_id()[1]
                chain1 = atom[2].get_id()

                atomname2 = metal[0].name
                resnum2 = metal[1].get_id()[1]
                chain2 = metal[2].get_id()

                atom_dist = atom[0] - metal[0]
                out = yaml_string.format(spring_const, atom_dist, chain1, resnum1, atomname1, chain2, resnum2,
                                         atomname2)

                output.append(out)

            output = list(set(output))

            if output:
                output = ['{}'.format(o) for o in output]

    return output


def main(original_constraints, protein_file, original_input, permissive=False, all_metals=False, external=None, logger=None):
    metals, structure = find_metals(protein_file)
    if external:
        external = map_constraints(protein_file, original_input, original_constraints)
    output = find_geometry(metals, structure, permissive, all_metals, external, logger=logger)
    return output, external



import pele_platform.Errors.custom_errors as ce


def check_limit_number_atoms(sdf: str, limit_atoms: int):
    from rdkit import Chem
    """ Function to check input file does not excit certain number of atoms """

    mols = Chem.SDMolSupplier(sdf)
    for mol in mols:
        atoms = mol.GetAtoms()
        natoms = len([atom for atom in atoms])
        if natoms > limit_atoms:
            name_ligand = mol.GetProp("_Name")
            raise ce.LigandSizeExceed(f"Ligand {name_ligand} in {sdf} exceeds size {limit_atoms}")

  
def check_substructure_match(ligand, ligand_core, core_atoms):
    import rdkit.Chem.rdMolAlign as rd
    from rdkit import Chem

    # Initialize ring information
    Chem.GetSSSR(ligand); Chem.GetSSSR(ligand_core)  

    # Align and obtain vocabulary between atoms in core and ligand
    align_obj = rd.GetO3A(ligand, ligand_core)
    atom_map = align_obj.Matches()
    vocabulary = { i2:i1 for i1, i2 in atom_map } 

    # Check the substructure search and change indexes if dont agree with alignment
    for idx_atom_ligand, idx_atom_core in enumerate(core_atoms):
        if idx_atom_ligand not in vocabulary: # If it's a removed hydrogen - pass
            continue
        if vocabulary[idx_atom_ligand] == idx_atom_core: # If indexes are correct - pass
            continue
        else: # If indexes are incorrect - swap indexes
            core_atoms = list(core_atoms)
            core_atoms[idx_atom_ligand] = vocabulary[idx_atom_ligand]
    return core_atoms



import pele_platform.constants.constants as cs

class InOutParams(object):


    def generate_inout_params(self, args):
        #If User inputs exit condition or when doing an exit simulation
        if args.exit or args.in_out or args.in_out_soft:
            self.exit_value = args.exit_value if args.exit_value else self.simulation_params.get("exit_value", 0.9)
            self.exit_condition = args.exit_condition if args.exit_condition else self.simulation_params.get("exit_condition", ">")
            self.exit_trajnum = args.exit_trajnum if args.exit_trajnum else self.simulation_params.get("exit_trajnum", 4)
            self.bias_column = args.bias_column if args.bias_column else self.simulation_params.get("bias_column", 6)
            self.unbinding_block = cs.UNBINDING.format(self.bias_column, self.exit_value, self.exit_condition, self.exit_trajnum)
            self.equilibration = "false" #Lots of problems look into it
        else:
            self.unbinding_block = ""



from Bio.PDB import PDBParser, PDBIO, NeighborSearch, Selection, rotaxis
from Bio.PDB.vectors import Vector
import numpy as np
import os, argparse
import time


def calculate_com(structure):
    structure_mass = 0.0
    com = np.zeros(3)
    for atom in structure.get_atoms():
        com = com + np.array(list(atom.get_vector())) * atom.mass
        structure_mass += atom.mass
    com = com / structure_mass
    return com


def randomize_starting_position(ligand_file, complex_file, outputfolder=".", nposes=200, test=False, user_center=None,
                                logger=None):
    """
    Randomize initial ligand position around the receptor.
    Default number of poses = 200.
    :param ligand_file:
    :param complex_file:
    :param nposes:
    :return:
    """
    if test:  np.random.seed(42)

    # read in files
    parser = PDBParser()
    output = []
    structure = parser.get_structure('protein', complex_file)
    ligand = parser.get_structure('ligand', ligand_file)
    COI = np.zeros(3)

    # get center of interface (if PPI)
    if user_center:
        chain_id, res_number, atom_name = user_center.split(":")
        for chain in structure.get_chains():
            if chain.id == chain_id:
                for residue in chain.get_residues():
                    if residue.id[1] == int(res_number):
                        for atom in residue.get_atoms():
                            if atom.name == atom_name: 
                                COI = np.array(list(atom.get_vector())) 
  
    # calculate protein and ligand COM
    com_protein = calculate_com(structure)
    com_ligand = calculate_com(ligand)

    # calculating the maximum d of the ligand
    coor_ligand = []
    for atom in ligand.get_atoms():
        coor_ligand.append(list(atom.get_vector() - com_ligand))

    coor_ligand = np.array(coor_ligand)
    coor_ligand_max = np.amax(coor_ligand, axis=0)
    d_ligand = np.sqrt(np.sum(coor_ligand_max ** 2))

    # set threshold for near and far contacts based on ligand d
    if d_ligand / 2 < 5.0:
        d5_ligand = 5.0
    else:
        d5_ligand = d_ligand / 2 + 1

    if d_ligand > 8.0:
        d8_ligand = d_ligand / 2 + 4
    else:
        d8_ligand = 8.0

    # calculate vector to move the ligandi
    if user_center:
        move_vector = com_ligand - COI
    else:
        move_vector = com_ligand - com_protein

    # translate the ligand to the protein COM (COI for PPI)
    original_coords = []
    for atom in ligand.get_atoms():
        ligand_origin = np.array(list(atom.get_vector())) - move_vector
        original_coords.append(ligand_origin)
        atom.set_coord(ligand_origin)

    # calculating the maximum radius of the protein from the origin
    coor = []
    for atom in structure.get_atoms():
        coor.append(list(atom.get_vector() - com_protein))
    coor = np.array(coor)
    coor_max = np.amax(coor, axis=0)
    d = np.sqrt(np.sum(coor_max ** 2))

    # radius of the sphere from the origin
    D = 10.0 if user_center else np.ceil(6.0 + d)
    D_initial = D
    logger.info("Sampling {}A spherical box around the centre of the receptor/interface.".format(D))

    if user_center:
        sphere_cent = COI
    else:
        sphere_cent = com_protein

    j = 0
    logger.info("Generating {} poses...".format(nposes))
    start_time = time.time()
    while (j < nposes):
        # generate random coordinates
        phi = np.random.uniform(0, 2 * np.pi)
        costheta = np.random.uniform(-1, 1)
        u = np.random.uniform(0, 1)
        theta = np.arccos(costheta)

        r = D * np.cbrt(u)
        x = r * np.sin(theta) * np.cos(phi)
        y = r * np.sin(theta) * np.sin(phi)
        z = r * np.cos(theta)

        # move ligand to the starting point (protein COM)
        for atom, coord in zip(ligand.get_atoms(), original_coords):
            atom.set_coord(coord)

        # translate ligand to a random position
        translation = (x, y, z)
        for atom in ligand.get_atoms():
            new_pos_lig_trans = np.array(list(atom.get_vector())) - translation
            atom.set_coord(new_pos_lig_trans)

        # calculate ligand COM in the new position
        new_ligand_COM = calculate_com(ligand)

        # rotate ligand
        vector = Vector(new_ligand_COM)
        rotation_matrix = rotaxis(np.random.randint(0, 2 * np.pi), vector)

        for atom in ligand.get_atoms():
            coords_after = atom.get_vector().left_multiply(rotation_matrix)
            atom.set_coord(coords_after)

        # check if it's inside the sampling sphere
        dist = np.sqrt((new_ligand_COM[0] - sphere_cent[0]) ** 2 + (new_ligand_COM[1] - sphere_cent[1]) ** 2 + (
                new_ligand_COM[2] - sphere_cent[2]) ** 2)

        if dist < D:
            # check contacts at: 5A (no contacts) and 8A (needs contacts)
            protein_list = Selection.unfold_entities(structure, "A")
            contacts5 = []
            contacts8 = []
            ligand_atoms = list(ligand.get_atoms())
            
            contacts5.append( NeighborSearch(protein_list).search(new_ligand_COM, d5_ligand, "S"))
            contacts8 = NeighborSearch(protein_list).search(new_ligand_COM, d8_ligand, "S")
            if contacts8 and not any(contacts5):
                j += 1
                io = PDBIO()
                io.set_structure(ligand)
                output_name = os.path.join(outputfolder, 'ligand{}.pdb'.format(j))
                io.save(output_name)
                output.append(output_name)
                start_time = time.time()

            end_time = time.time()
            total_time = end_time - start_time
            if total_time > 60:
                D += 1
                if D - D_initial >= 20:
                    logger.info("Original box increased by 20A. Aborting...")
                    break
                start_time = end_time
                logger.info("Increasing sampling box by 1A.")
    logger.info("{} poses created successfully.".format(j))
    return output, D, list(sphere_cent)


def join(receptor, ligands, residue, output_folder=".", output="input{}.pdb"):
    """
    Join receptor and ligand PDB files into one, conserving old formatting
    and not repeating atom numbers.
    """

    with open(receptor, "r") as f:
        lines = f.readlines()
        receptor_content = [line for line in lines if
                            line[17:20] != residue and not line.startswith("CONECT") and not line.startswith("END")]
        connects = [line for line in lines if line.startswith("CONECT")]
        ligand_content_without_coords = [line[0:27] + "{}" + line[56:] for line in lines if line[17:20] == residue]
        atom_nums = [line[6:11] for line in lines if line[17:20] == residue]

    outputs = []
    for i, ligand in enumerate(ligands):
        with open(ligand, "r") as fin:
            # exclude connects but keep initial atom names (CL problem)
            ligand_coords = {line[12:16].strip(): line[27:56] for line in fin if
                             line.startswith("ATOM") or line.startswith("HETATM")}
            # assert len(ligand_coords) == len(ligand_content_without_coords), "Experimental part - send an issue to github"

            ligand_content = []
            for pdb_block in ligand_content_without_coords:
                atom_name = pdb_block[12:16].strip()
                coord = ligand_coords[atom_name]
                ligand_pdb_line = pdb_block.format(coord)
                ligand_content.append(ligand_pdb_line)

            for j, line in enumerate(ligand_content):
                ligand_content[j] = line[:6] + "{:>5}".format(atom_nums[j]) + line[11:]
        content_join_file = receptor_content + ["TER\n"] + ligand_content + ["TER\n"] + connects + ["END"]
        output_path = os.path.join(output_folder, output.format(i))
        with open(output_path, "w") as fout:
            fout.write("".join(content_join_file))
        outputs.append(output_path)

    return outputs


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--ligand", type=str, required=True, help="Ligand pdb file")
    parser.add_argument("--receptor", type=str, required=True, help="Receptor pdb file")
    parser.add_argument("--resname", type=str, required=True, help="Ligand resname")
    parser.add_argument("--poses", type=int, default=20, help="How many input poses to produce")
    parser.add_argument("--output_folder", type=str, default=".", help="output folder")
    parser.add_argument("--user_center", type=str, default=None,
                        help="Center of protein-protein interface (chain:residue number:atom name)")
    args = parser.parse_args()
    return os.path.abspath(args.ligand), os.path.abspath(
        args.receptor), args.resname, args.poses, args.output_folder, args.user_center


if __name__ == "__main__":
    ligand, receptor, resname, poses, output_folder, user_center = parse_args()
    output, sphere_cent, D = randomize_starting_position(ligand, receptor, output_folder, poses, None, user_center)
    # Make format ready to be captured"
    print("OUTPUT; {}. Sphere center: {}. D: {}".format(output, sphere_cent, D))



'''
Created on Oct 28, 2013

@author: israel

Script to generate OBC parameters for Heteroatoms. The parameters have been extracted from Tinker Molecular package.
This script uses the impact template generated for PELE to identify the atom type and assign the right parameters for the OBC solvent.
If one parameter have not been defined in the OBC templates it puts the default parameter and prints a warning.

Usage:

python solventOBCParamsGenerator.py ALA  

Results:
The ouput will be ALA_OBCParams.txt in the same folder of ALA. Then, to put these new parameters in PELE you have to copy
the data inside the file in the pele template. (Data/OBC/solventParamsHCTOBC.txt) and that's all.

'''

import sys


paramtersLst =[ ['CW', '1.875', '0.72'] ,
                ['NC', '1.7063', '0.79'] ,
                ['CM', '1.875', '0.72'] ,
                ['C*', '1.875', '0.72'] ,
                ['H1', '1.25', '0.85'] ,
                ['CT', '1.9', '0.72'] ,
                ['N2', '1.7063', '0.79'] ,
                ['N*', '1.7063', '0.79'] ,
                ['CR', '1.875', '0.72'] ,
                ['HO', '1.05', '0.85'] ,
                ['NB', '1.7063', '0.79'] ,
                ['H2', '1.25', '0.85'] ,
                ['S', '1.775', '0.96'] ,
                ['NA', '1.7063', '0.79'] ,
                ['H4', '1.25', '0.85'] ,
                ['HC', '1.25', '0.85'] ,
                ['C', '1.875', '0.72'] ,
                ['OH', '1.535', '0.85'] ,
                ['CQ', '1.875', '0.72'] ,
                ['CK', '1.875', '0.72'] ,
                ['O2', '1.48', '0.85'] ,
                ['OS', '1.535', '0.85'] ,
                ['SH', '1.775', '0.96'] ,
                ['HA', '1.25', '0.85'] ,
                ['CB', '1.875', '0.72'] ,
                ['H5', '1.25', '0.85'] ,
                ['CN', '1.875', '0.72'] ,
                ['P', '1.87', '0.86'] ,
                ['N3', '1.625', '0.79'] ,
                ['HP', '1.25', '0.85'] ,
                ['N', '1.7063', '0.79'] ,
                ['H', '1.15', '0.85'] ,
                ['HS', '1.25', '0.85'] ,
                ['CV', '1.875', '0.72'] ,
                ['CA', '1.875', '0.72'] ,
                ['O', '1.48', '0.85'] ,
                ['CC', '1.875', '0.72'] ,]
              #  ['HWS', '1.05', '0.85'] ,
              #  ['OWS', '1.535', '0.85'] ,]

# Parameter List extracted from TINKER source code for OBC  (ksolv.f: +/-line 423)
atomTypesOverlapFactors = [['H','1.25'],
                           ['Li', '1.432'] ,
                           ['C','1.90'],
                           ['N','1.7063'],
                           ['O','1.535'],
                    ['F', '1.47'] ,
                    ['FE', '2.00'] ,   #default parameters
                    ['Ne', '1.39'] ,
                    ['Na', '1.992'] ,
                    ['Mg', '1.70'] ,
                    ['Si', '1.80'] ,
                    ['P', '1.87'] ,
                    ['S', '1.775'] ,
                    ['Cl', '1.735'] ,
                    ['Ar', '1.70'] ,
                    ['K', '2.123'] ,
                    ['Ca', '1.817'] ,
                    ['Br', '1.90'] ,
                    ['Kr', '1.812'] ,
                    ['Rb', '2.26'] ,
                    ['I', '2.10'] ,
                    ['Xe', '1.967'] ,
                    ['Cs', '2.507'] ,
                    ['Ba', '2.188'] ,
                    ['Pt', '2.0'] ,
                  ]

# Parameter List extracted from TINKER source code for OBC  (ksolv.f: +/-line 423)
atomTypesHCTradii = [['H','0.85'],
                     ['C','0.72'],
                     ['N','0.79'],
                     ['O','0.85'],
                     ['F','0.88'],
                  ['P','0.86'],
                  ['S','0.96'],
                  ['Pt','0.80'],
                  ['FE','0.88'],
                  ]


# Parse the impact template to get the residue name and the atom names and atom types
def parseImpactTemplate(impactTemplate):
    
    try:
        fileout = open(impactTemplate)
    except OSError:
        env.logger.info('Impossible to open the file ....  ',impactTemplate)
        sys.exit()
    
    name = ''
    atomNamesAndTypes = []
    atomNumber = []    

    read = False
    bonds = []

    for line in fileout:
        if line[0]!=' ' and len(line.split())==6:
            name = line.split()[0].upper()
            
        if len(line)== 68 and len(line.split())==9:
            tmp = line.split()
            atomNamesAndTypes.append([tmp[4].replace('_',''),tmp[3]])
            atomNumber.append(tmp[0])           

        if 'BOND' in line[:5]: 
            read = True
            continue    
        if 'THET' in line[:5]:
            read = False    
            continue
        if read:
            tmp2 = line.split()
            bonds.append([tmp2[0],tmp2[1]]) 

    # get number of atoms connected
 
    numberOfConnections = []


    for ele in atomNumber:
        counter = 0
        atomAttached = ''
        for bond in bonds:
            if ele==bond[0]:
                counter +=1
                if getShortName(atomNamesAndTypes[int(ele[0])-1][0])[0]=='H':
                    atomAttached= getShortName(atomNamesAndTypes[int(bond[1])-1][0])[0]
            if ele==bond[1]:
                counter +=1
                if getShortName(atomNamesAndTypes[int(ele[0])-1][0])[0]=='H':
                    atomAttached= getShortName(atomNamesAndTypes[int(bond[0])-1][0])[0]

        numberOfConnections.append([counter,atomAttached])

 
    return name, atomNamesAndTypes,numberOfConnections


# Assign overlapFactors and HCT radii using the atom name
def getOverlapscalefactorsFromAtomName(atomName,atomTypesOverlapscalefactors,atomTypesHCTradiinum,numberOfBonds):
   
    shortName = getShortName(atomName)

    radii ='0.80'
    found = False

    overlapFactor, found = assignOverlapFactor(shortName,atomTypesOverlapscalefactors,numberOfBonds,atomName)

    if not found:
        shortName = shortName[0]

        overlapFactor,found = assignOverlapFactor(shortName,atomTypesOverlapscalefactors,numberOfBonds,atomName)

    for atomType in atomTypesHCTradii:
        if shortName==atomType[0].upper():
            radii = atomType[1]

    return overlapFactor,radii,found

def assignOverlapFactor(name,atomTypeOverlapFactorTable,numberOfBonds,realName):
    
    overlapFactor = ''

    found = False

    for atomType in atomTypeOverlapFactorTable:
            if name==atomType[0].upper():
                overlapFactor = atomType[1]
                found = True

    if name=='H' and numberOfBonds[1]=='O': 
        overlapFactor='1.05' 
    if name=='H' and numberOfBonds[1]=='N': overlapFactor='1.15' 

    if name=='C' and numberOfBonds[0]==3: overlapFactor='1.875' 
    if name=='C' and numberOfBonds[0]==2: overlapFactor='1.825' 
    if name=='N' and numberOfBonds[0]==4: overlapFactor='1.625' 
    if name=='N' and numberOfBonds[0]==1: overlapFactor='1.60' 
    if name=='O' and numberOfBonds[0]==1: overlapFactor='1.48' 
 
    return overlapFactor,found

# extract atom type from atom name for hetero atoms
def getShortName(name):
#    shortName = ''.join(i for i in name if not i.isdigit()).upper()

    size = len(name)
    start = 0
    end = 0

    for iterator in range(0,size):
        char = name[iterator]
        if not char.isdigit():
            start = iterator
            break   

    for iterator in reversed(range(0,size)):
        char = name[iterator]
        if not char.isdigit():
            end = iterator
            break   

    short = name[start:end+1]

    return short


#generate the final output template
def generateSolventTemplate(residueName, atomNamesAndTypes,paramtersLst,atomTypesOverlapFactors,atomTypesHCTradii,numberOfBonds):
    
    found = False
    foundNoStd = False
    
    templateParameters = []
    
    for i,ele in enumerate(atomNamesAndTypes):
        found = False
        for param in paramtersLst:
            if ele[1]==param[0]:
                templateParameters.append(residueName+'Z   '+ele[0]+'   '+ele[1]+'    '+param[1]+'   '+param[2]+'\n') 
                found = True
                
        if not found:
            foundNoStd = False
            
            overlapFactor, radii, foundNoStd = getOverlapscalefactorsFromAtomName(ele[0],atomTypesOverlapFactors,atomTypesHCTradii,numberOfBonds[i])
            
            if not foundNoStd:
                env.logger.info('Parameter NOT found in the template database ....... '+ele[0]+'   '+ele[1]+'  using default parameters')

            templateParameters.append(residueName+'Z   '+ele[0]+'   '+ele[1]+'    '+overlapFactor+'   '+radii+'\n') 
        
    return templateParameters

def main(template, out_file):
    # Parameter List extracted from PELE solvent templates for OBC 
    residueName, atomNamesAndTypes, numberOfBonds = parseImpactTemplate(template)
    
    templateParameters =generateSolventTemplate(residueName, atomNamesAndTypes,paramtersLst,atomTypesOverlapFactors,atomTypesHCTradii,numberOfBonds)
    
    fileout = open(out_file, 'a')
    fileout.writelines([item for item in templateParameters]) 
    fileout.close()
    
    





class PPIParams(object):

    def generate_ppi_params(self, args):
        self.ppi = args.ppi
        self.center_of_interface = args.center_of_interface
        self.n_components = args.n_components if args.n_components else self.simulation_params.get("n_components", 25)
